#ifdef _WIN32
#include <windows.h>
#include <process.h>  // _beginthread, _endthread 
#else
#include "wintypes.h"
#ifdef USE_GTK
#include <gtk/gtk.h>
#endif
#endif
#include <stdio.h>
#include <GL/gl.h>
//#include <glext.h>
#include <GL/glu.h>

#include "gfx.h"
#ifdef _WIN32
#include "wingl.h"
#else
#include "sdlgl.h"
#endif
#include "crc_ucode.h"

//** externals
void RealExecuteDList(unsigned __int32 dwAddr);
void RefreshVisual(void);
void ResizeVisualClipRectangle();
int GetUCode();
void ClearUCode();
void DelTexture();
void init_cache();

//** defines
#define GFX_VERSION					0x0102

//** globals
BOOLEAN	 bFullScreen; // FullScreen Mode

int				*pInterruptMask;
unsigned char	*pRDRAM;
unsigned char	*pIMEM;
unsigned char	*pDMEM;
unsigned char	*pVIREG;
unsigned char	*pROM;
_u32			*pVIORG;
_u32			*pVISTS;

//BOOL			Draw2d = FALSE;
BOOL			Draw2d = TRUE;
BOOL			NoBlending = FALSE;			
BOOL			m_bFullScreen = FALSE;
#ifdef _WIN32
RECT			m_WindowRect;
WINDOWPLACEMENT m_WindowPlacement;
HMENU			m_hOldMenu;
DWORD			m_iOldWindowStyle;
#endif

int				ucode_version;
int skipCount;

#ifdef _WIN32
HINSTANCE g_hInstance;
HWND g_hWnd;

DEVMODEA m_DMsaved,devMode;

extern WINDATA_T WinData;

extern int ucode;
#else
int ucode;
BOOL FogEnabled = FALSE;
BOOL FrameBuffer = FALSE;
#endif // _WIN32
extern BOOL HandsetUcode;
extern float fScalex;
extern float fScaley;

float imgHeight = 240.0f;
float imgWidth  = 320.0f;

#ifdef _WIN32

#define V_DONOTHING 0
#define V_EXEDLIST 1
#define V_CLOSEROM 2
#define V_STARTROM 3

HANDLE  hVideoThread;
BOOL VideoDone = FALSE;
int VideoDoNow = V_DONOTHING;
BOOL VideoBusy = FALSE;
#endif // _WIN32

void rdp_reset();

#ifdef _WIN32
void VideoThread()
{
	WGL_Init();
    SetThreadPriority(hVideoThread, THREAD_PRIORITY_HIGHEST);

	while(!VideoDone)
	{
		switch (VideoDoNow)
		{
		case V_DONOTHING:
			break;
		case V_EXEDLIST:
			{
				unsigned __int32 addr = ((unsigned __int32 *)pDMEM)[0xff0>>2];

				RECT clRect;
				LPRECT lpRect = &clRect;
				GetClientRect(WinData.hWnd,lpRect);

				WinData.Width = clRect.right-clRect.left;
				WinData.Height = clRect.bottom-clRect.top;;

				fScalex = WinData.Width / imgWidth;
				fScaley = WinData.Height / imgHeight;

//				glOrtho(0, WinData.Width, WinData.Height, 0, -1, 1);
				glOrtho(0, WinData.Width, WinData.Height, 0, -0, 1023);
				ResizeVisualClipRectangle();

				ucode_version = GetUCode();
				
				VideoBusy = TRUE;
				RealExecuteDList(addr);
			}
			break;
		case V_CLOSEROM:
			DelTexture();
			//WGL_UnInit();
			break;
		case V_STARTROM:
			ClearUCode();
			//WGL_Init();
			break;
		}

		VideoBusy = FALSE;
		SuspendThread(hVideoThread);
	}

	_endthreadex(1);
}
#endif
/*
===============================================================================

  All our exported functions...

===============================================================================
*/
void SaveUCode(int _ucode);

#ifdef _WIN32
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD dwReason, LPVOID lpvReserved)
{
	ucode = -1;
	HandsetUcode = FALSE;

	g_hInstance = hinstDLL;


	return TRUE;
}

//void   (__cdecl* DllClose)(void) = NULL;

void   __cdecl DllClose (void)
{
	VideoDone = TRUE;
	ResumeThread(hVideoThread);
}

// __________________________________________________________________________________________________
// SetWindowMode
//
int 
SetWindowedMode(int _width, int _height, int _bpp)
{



    //SetWindowLong(WinData.hWnd, GWL_STYLE, m_iWindowStyle);
    SetWindowLong(WinData.hWnd, GWL_STYLE, 0);
	ShowWindow(WinData.hWnd, SW_SHOWNORMAL);
	
	if (!ChangeDisplaySettings(NULL,CDS_TEST)) { 	// If The Shortcut Doesn't Work
		ChangeDisplaySettings(NULL,CDS_RESET);		// Do It Anyway (To Get The Values Out Of The Registry)
		ChangeDisplaySettings(&m_DMsaved,CDS_RESET);	// Change It To The Saved Settings
	} else {
		ChangeDisplaySettings(NULL,CDS_RESET);		// If It Works, Go Right Ahead
	}

	//
	// Reinit Winmdow
	if (WinData.hStatusBar) { ShowWindow(WinData.hStatusBar,SW_SHOW); }
	if (m_hOldMenu) { 
		SetMenu(WinData.hWnd,m_hOldMenu); 
		m_hOldMenu = NULL;
	}
	SetWindowLong(WinData.hWnd, GWL_STYLE, m_iOldWindowStyle);
	SetWindowPos(WinData.hWnd, HWND_NOTOPMOST, m_WindowPlacement.rcNormalPosition.left, 
		m_WindowPlacement.rcNormalPosition.top, 640, 480, SWP_NOSIZE|SWP_SHOWWINDOW);
	ShowCursor(TRUE);
	
	m_bFullScreen = FALSE;

	SetWindowPos(WinData.hWnd, HWND_TOP, m_WindowRect.left, m_WindowRect.top, 
							m_WindowRect.right - m_WindowRect.left,
							m_WindowRect.bottom - m_WindowRect.top,
							SWP_SHOWWINDOW);

	GetWindowRect(WinData.hWnd, &m_WindowRect);

//	_trace_("COpenGL: Window-Mode\n");
    return(FALSE);
}


// __________________________________________________________________________________________________
// SetFullscreenMode
//
int SetFullscreenMode(int _width, int _height, int _bpp);

int SetFullscreenMode(int _width, int _height, int _bpp)
{

	ZeroMemory(&devMode, sizeof(DEVMODE));
	
	devMode.dmSize       = sizeof(DEVMODE);
	devMode.dmPelsWidth	 = _width;
	devMode.dmPelsHeight = _height;
	devMode.dmBitsPerPel = _bpp;
	devMode.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;
	
	if (ChangeDisplaySettings(&devMode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
	{
		m_bFullScreen = FALSE;
		return FALSE;
	}

	//
	// dont show the mouse cursor in fullscreen
	ShowCursor(0);

	//
	// ReInit the Window
	m_WindowPlacement.length = sizeof(m_WindowPlacement);
	GetWindowPlacement(WinData.hWnd, &m_WindowPlacement);
	if (WinData.hStatusBar) 
	{ 
		ShowWindow(WinData.hStatusBar,SW_HIDE); 
	}
	m_hOldMenu= GetMenu(WinData.hWnd);
	if (m_hOldMenu) 
	{ 
		SetMenu(WinData.hWnd,NULL); 
	}

	m_iOldWindowStyle = GetWindowLong(WinData.hWnd,GWL_STYLE);
	SetWindowLong(WinData.hWnd, GWL_EXSTYLE, WS_EX_APPWINDOW | WS_EX_TOPMOST);
	SetWindowLong(WinData.hWnd, GWL_STYLE, WS_VISIBLE|WS_POPUP);
	
	SetWindowPos(WinData.hWnd, HWND_TOPMOST, 0, 0, _width,_height, SWP_SHOWWINDOW);


	//
	// setup
	m_bFullScreen = TRUE;
//	m_iFullscreenWidth = _width;
//	m_iFullscreenHeight = _height;


//	_trace3_("COpenGL: Fullscreen set to width: %i height: %i bpp: %i\n", _width, _height, _bpp);
	return TRUE;
}

void ChangeWinSize()
{
}
#endif // _WIN32
//EXPORT void CALL CloseDLL (void);
void   __cdecl CloseDLL (void)
{
#ifdef _WIN32
	SaveUCode(ucode_version);
//	WGL_UnInit();
#endif
}

//void   (__cdecl* ChangeWindow)(BOOL) = NULL;

void   __cdecl ChangeWindow (void)
{
#ifdef _WIN32
    static m_bFullScreen = 0;

    if (m_bFullScreen == 0) 
    {
    	EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &m_DMsaved);
                
        GetWindowRect(WinData.hWnd, &m_WindowRect);    // remember windowed mode's rect for later
//        ChangeWinSize(WinData.hWnd, m_DMsaved.dmPelsWidth, m_DMsaved.dmPelsHeight, NULL);
        m_bFullScreen = SetFullscreenMode(m_DMsaved.dmPelsWidth, m_DMsaved.dmPelsHeight, 16);
    }
    else
    {
//        ChangeWinSize(WinData.hWnd, 640, 480, NULL);
        m_bFullScreen = SetWindowedMode(640, 480, 16);
    }
#else
    SDLGL_ToggleFullScreen();
#endif
}

//void   (__cdecl* DllAbout)(HWND) = NULL;

char plgname[] = "TR64 OpenGL";
int vermjr = 0;
int vermnr = 7;
char versfx[] = ".8";

void   __cdecl DllAbout ( HWND hParent )
{
#ifdef _WIN32
	char tMsg[256];
	char tTilte[256];

	sprintf(tTilte,"%s v%i.%i%s - About Box",plgname,vermjr,vermnr,versfx);
	sprintf(tMsg,"%s v%i.%i%s\nby Icepir8\ntexture cache modified by Hacktarux",plgname,vermjr,vermnr,versfx);

	MessageBox(NULL, tMsg, tTilte, MB_OK|MB_ICONINFORMATION);
#else
#ifdef USE_GTK
   char tMsg[256];
   GtkWidget *dialog, *label, *okay_button;
   
   dialog = gtk_dialog_new();
   sprintf(tMsg,"New TR64 OpenGL v%i.%i%s\nby Icepir8\ntexture cache modified by Hacktarux\noriginaly ported to linux by Blight",vermjr,vermnr,versfx);
   label = gtk_label_new(tMsg);
   okay_button = gtk_button_new_with_label("OK");
   
   gtk_signal_connect_object(GTK_OBJECT(okay_button), "clicked",
			     GTK_SIGNAL_FUNC(gtk_widget_destroy),
			     GTK_OBJECT(dialog));
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
		     okay_button);
   
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
		     label);
   gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
   gtk_widget_show_all(dialog);
#else
   char tMsg[256];
   sprintf(tMsg,"New TR64 OpenGL v%i.%i%s\nby Icepir8\ntexture cache modified by Hacktarux\noriginaly ported to linux by Blight",vermjr,vermnr,versfx);
   fprintf(stderr, "TR64GL: About\n%s\n", tMsg);
#endif
#endif
}

//void   (__cdecl* DllConfig)(HWND) = NULL;

void   __cdecl DllConfig ( HWND hParent )
{
#ifdef _WIN32
	CreateConfigDlg();
#else
        fprintf(stderr, "TR64GL: No Config dialog!\n");
#endif
}

//void   (__cdecl* DllTest)(HWND) = NULL;

void   __cdecl DllTest ( HWND hParent )
{
#ifdef _WIN32
	MessageBox(NULL, "No Test Box yet", "message", MB_OK);
#else
        fprintf(stderr, "TR64GL: No test yet!\n");
#endif
}

//void   (__cdecl* DrawScreen)(void) = NULL;

void   __cdecl DrawScreen (void)
{

}

//void   (__cdecl* GetDllInfo )( PLUGIN_INFO *) = NULL;

void   __cdecl GetDllInfo ( PLUGIN_INFO * PluginInfo )
{	
	PluginInfo->NormalMemory = FALSE;
	PluginInfo->MemoryBswaped = TRUE;
	PluginInfo->Version = GFX_VERSION;
	PluginInfo->Type = PLUGIN_TYPE_GFX;
	sprintf(PluginInfo->Name,"%s v%i.%i%s",plgname,vermjr,vermnr,versfx);
}

//#define USE_THREADS

void (*CheckInterrupts)( void );

//BOOL   (__cdecl* InitiateGFX)(GFX_INFO) = NULL;
BOOL interlaced = FALSE;

BOOL   __cdecl InitiateGFX (GFX_INFO Gfx_Info)
{
#ifdef _WIN32
	RECT clRect;
	LPRECT lpRect = &clRect;
#endif

	pInterruptMask = (int*)Gfx_Info.MI_INTR_REG;
	pRDRAM = Gfx_Info.RDRAM;
	pDMEM = Gfx_Info.DMEM;
	pIMEM = Gfx_Info.IMEM;
	pROM = Gfx_Info.HEADER;
	pVIREG = (unsigned char*)Gfx_Info.DPC_START_REG;
	pVIORG = (_u32*)Gfx_Info.VI_ORIGIN_REG;
	pVISTS = (_u32*)Gfx_Info.VI_STATUS_REG;
	interlaced = ((*pVIORG & 0x040) != 0);
	CheckInterrupts = Gfx_Info.CheckInterrupts;

#ifdef _WIN32
//	WinData.Width = 640;
//	WinData.Height = 480;
	WinData.hWnd = Gfx_Info.hWnd;
	WinData.hStatusBar = Gfx_Info.hStatusBar;

	GetClientRect(WinData.hWnd,lpRect);

	WinData.Width = clRect.right-clRect.left;
	WinData.Height = clRect.bottom-clRect.top;;

	fScalex = WinData.Width / imgWidth;
	fScaley = WinData.Height / imgHeight;

//	if (interlaced) fScaley /= 2.0f;

//	if(!WGL_Init()) return FALSE;
#ifdef USE_THREADS
    hVideoThread = (HANDLE)_beginthread((void*)VideoThread, 4096, NULL);
#else //USE_THREADS
//	WGL_Init();
#endif //USE_THREADS
#else
        // default configuration
        window.width = 640;
	window.height = 480;
	window.ubpp = 32;
	window.flags.gleGL_POINT_SMOOTH_HINT = GL_DONT_CARE;
	window.flags.gleGL_LINE_SMOOTH_HINT = GL_DONT_CARE;
	window.flags.gleGL_POLYGON_SMOOTH_HINT = GL_DONT_CARE;
	window.flags.gleGL_PERSPECTIVE_CORRECTION_HINT = GL_DONT_CARE;
	fScalex = window.width / 320.0f;
	fScaley = window.height / 240.0f;
#endif
	return TRUE;
}

//void   (__cdecl* MoveScreen)(int, int) = NULL;

void   __cdecl MoveScreen (int xpos, int ypos)
{
}

//void   (__cdecl* ProcessDList)(void) = NULL;
void   __cdecl ProcessDList(void)
{
	unsigned __int32 DlistAddr = ((unsigned __int32 *)pDMEM)[0xff0>>2];

#ifdef USE_THREADS    
	VideoBusy = TRUE;
	VideoDoNow = V_EXEDLIST;
	
	ResumeThread(hVideoThread);

	while(VideoBusy)
	{
		Sleep(1);
	}
#else // USE_THREADS
	{
		unsigned __int32 addr = ((unsigned __int32 *)pDMEM)[0xff0>>2];
#ifdef _WIN32
		RECT clRect;
		LPRECT lpRect = &clRect;
		GetClientRect(WinData.hWnd,lpRect);

		WinData.Width = clRect.right-clRect.left;
		WinData.Height = clRect.bottom-clRect.top;

//		if(rdp_reg.colorimg_width > 100)
		{
//			imgWidth = rdp_reg.colorimg_width;
			imgHeight = (imgWidth / 4.0f) * 3.0f;
		}

		interlaced = ((*pVIORG & 0x040) != 0);

		fScalex = WinData.Width / imgWidth;
		fScaley = WinData.Height / imgHeight;

//		if (interlaced) fScaley /= 2.0f;

//		glOrtho(0, WinData.Width, WinData.Height, 0, -1, 1);
		glOrtho(0, WinData.Width, WinData.Height, 0, -0, 1023);
		ResizeVisualClipRectangle();

		ucode_version = GetUCode();
				
		VideoBusy = TRUE;
		RealExecuteDList(addr);
		CheckInterrupts();
#else // _WIN32
	        imgHeight = (imgWidth / 4.0f) * 3.0f;
	        interlaced = ((*pVIORG & 0x040) != 0);
	   
	        fScalex = window.width / imgWidth;
	        fScaley = window.height / imgHeight;
	   
	        //glOrtho(0, window.width, window.height, 0, -1, 1);
	        glOrtho(0, window.width, window.height, 0, -0, 1023);
	        ResizeVisualClipRectangle();
	   
	        RealExecuteDList(addr);
	        CheckInterrupts();
#endif // _WIN32
	}
#endif // USE_THREADS
}

void   __cdecl ProcessRDPList(void)
{
}

//void   (__cdecl* RomClosed)(void) = NULL;
extern HWND OglWindow;

void   __cdecl RomClosed (void)
{
#ifdef _WIN32
	SaveUCode(ucode_version);
#ifdef USE_THREADS
	ResumeThread(hVideoThread);
	VideoDoNow = V_CLOSEROM;
	while(VideoBusy)
	{
		Sleep(1);
	}
	ResumeThread(hVideoThread);
#else
	__try
	{
		DelTexture();
		rdp_reset();
	}
	__except(NULL, EXCEPTION_EXECUTE_HANDLER)
	{
	}
#endif // USE_THREADS
//	if (OglWindow)
//		EndDialog(OglWindow, TRUE);
#else // _WIN32
        rdp_reset();
        ucode = -1;
        DelTexture();
        SDLGL_UnInit();
#endif // _WIN32
}

//void   (__cdecl* UpdateScreen)(void) = NULL;

void   __cdecl UpdateScreen (void)
{
}

//void   (__cdecl* ViStatusChanged)(void) = NULL;

void   __cdecl ViStatusChanged (void)
{
}

//void   (__cdecl* ViWidthChanged)(void) = NULL;

void   __cdecl ViWidthChanged (void)
{
}

#ifdef USE_GTK
static volatile int ucode_choosen;
static gint delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
      return TRUE;
}

static void validate_ucode(GtkWidget *widget, gpointer data)
{
      ucode_choosen = 1;
      gtk_widget_destroy(data);
}

static void ucode0_clicked(GtkWidget *widget, gpointer data)
{
      ucode_version = 0;
}

static void ucode1_clicked(GtkWidget *widget, gpointer data)
{
      ucode_version = 1;
}

static void ucode2_clicked(GtkWidget *widget, gpointer data)
{
      ucode_version = 2;
}

static void ucode3_clicked(GtkWidget *widget, gpointer data)
{
      ucode_version = 3;
}

static void ucode4_clicked(GtkWidget *widget, gpointer data)
{
      ucode_version = 4;
}

static void ucode5_clicked(GtkWidget *widget, gpointer data)
{  
      ucode_version = 5;
}

static void ucode6_clicked(GtkWidget *widget, gpointer data)
{
      ucode_version = 6;
}

extern BOOL FrameBuffer;
static void frame_clicked(GtkWidget *widget, gpointer data)
{
      FrameBuffer = 1 - FrameBuffer;
}

extern BOOL FogEnabled;
static void fog_clicked(GtkWidget *widget, gpointer data)
{
      FogEnabled = 1 - FogEnabled;
}
#endif // USE_GTK

//void   (__cdecl* RomOpen)(void) = NULL;

void   __cdecl RomOpen (void)
{
#ifdef _WIN32
	init_cache();
	ClearUCode();

#ifdef USE_THREADS
	VideoDoNow = V_STARTROM;
	while(VideoBusy)
	{
		Sleep(1);
	}
	ResumeThread(hVideoThread);
//	if(!WGL_Init()) return;
#else // USE_THREADS
	WGL_Init();
#endif // USE_THREADS
#else  // _WIN32
#ifdef USE_GTK // _WIN32
   GtkWidget *dialog, *okay_button, *frame, *container, *note, *buffer, *fog;
   GtkWidget *ucode0, *ucode1, *ucode2, *ucode3, *ucode4, *ucode5, *ucode6;
   
   init_cache();
   ucode = -1;
   ucode_version = 0;
   dialog = gtk_dialog_new();
   okay_button = gtk_button_new_with_label("OK");
   frame = gtk_frame_new("Select UCode ...");
   container = gtk_vbox_new(FALSE, 0);
   ucode0 = gtk_radio_button_new_with_label(NULL, "0 - Ex. Mario64, Demos");
   ucode1 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(ucode0)),
					    "1 - Ex. Mario Kart");
   ucode2 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(ucode0)),
					    "2 - Ex. Waveracer USA");
   ucode3 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(ucode0)),
					    "3 - Ex. Mortal Kombat 4");
   ucode4 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(ucode0)),
					    "4 - Ex. Zelda");
   ucode5 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(ucode0)),
					    "5 - Ex. Perfect Dark");
   ucode6 = gtk_radio_button_new_with_label(gtk_radio_button_group(GTK_RADIO_BUTTON(ucode0)),
					    "6 - Ex. Diddy Kong Racing");
   note = gtk_label_new("Note: if this plugin crashes you probably\nhave selected a wrong UCode");
   buffer = gtk_check_button_new_with_label("Enable Frame Buffer Emulation");
   fog = gtk_check_button_new_with_label("Enable Fog Emulation");
   
   gtk_signal_connect(GTK_OBJECT(okay_button), "clicked",
		      GTK_SIGNAL_FUNC(validate_ucode), GTK_OBJECT(dialog));
   gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
		      GTK_SIGNAL_FUNC(delete_event), GTK_OBJECT(dialog));
   gtk_signal_connect(GTK_OBJECT(ucode0), "pressed",
		      GTK_SIGNAL_FUNC(ucode0_clicked), NULL);
   gtk_signal_connect(GTK_OBJECT(ucode1), "pressed",
		      GTK_SIGNAL_FUNC(ucode1_clicked), NULL);
   gtk_signal_connect(GTK_OBJECT(ucode2), "pressed",
		      GTK_SIGNAL_FUNC(ucode2_clicked), NULL);
   gtk_signal_connect(GTK_OBJECT(ucode3), "pressed",
		      GTK_SIGNAL_FUNC(ucode3_clicked), NULL);
   gtk_signal_connect(GTK_OBJECT(ucode4), "pressed",
		      GTK_SIGNAL_FUNC(ucode4_clicked), NULL);
   gtk_signal_connect(GTK_OBJECT(ucode5), "pressed",
		      GTK_SIGNAL_FUNC(ucode5_clicked), NULL);
   gtk_signal_connect(GTK_OBJECT(ucode6), "pressed",
		      GTK_SIGNAL_FUNC(ucode6_clicked), NULL);
   gtk_signal_connect(GTK_OBJECT(buffer), "pressed",
		      GTK_SIGNAL_FUNC(frame_clicked), NULL);
   gtk_signal_connect(GTK_OBJECT(buffer), "pressed",
		      GTK_SIGNAL_FUNC(fog_clicked), NULL);
   
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), frame);
   gtk_container_add(GTK_CONTAINER(frame), container);
   gtk_container_add(GTK_CONTAINER(container), ucode0);
   gtk_container_add(GTK_CONTAINER(container), ucode1);
   gtk_container_add(GTK_CONTAINER(container), ucode2);
   gtk_container_add(GTK_CONTAINER(container), ucode3);
   gtk_container_add(GTK_CONTAINER(container), ucode4);
   gtk_container_add(GTK_CONTAINER(container), ucode5);
   gtk_container_add(GTK_CONTAINER(container), ucode6);
   gtk_container_add(GTK_CONTAINER(container), note);
   gtk_container_add(GTK_CONTAINER(container), buffer);
   gtk_container_add(GTK_CONTAINER(container), fog);
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), okay_button);
   
   gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
   ucode_choosen = 0;
   gtk_widget_show_all(dialog);
   gtk_widget_show_now(dialog);
   
   while(!ucode_choosen);
   
   SDLGL_Init();
#else // USE_GTK
   char c, s[2];
   SDLGL_Init();
   init_cache();
   ucode = -1;
   ucode_version = -1;
   while (ucode_version < 0 || ucode_version > 6)
     {
	printf("please select the ucode:");
	c = getchar();
	s[0] = c;
	s[1] = 0;
	ucode_version = atoi(s);
     }
   printf("selected ucode : %d\n", ucode_version);
   SDLGL_Init();
#endif
#endif // _WIN32
}

void   __cdecl DLL_ResetGFX (void)
{
}
