/**
 * TrueReality - wingl.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
**/

#include "wintypes.h"
#include <malloc.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "sdlgl.h"
#include "debug.h"
//#include "trlogo.h"
#include "tricon.h"
#include "SDL.h"

TWindow window;
const char szAppName[] = "New TR64GL v0.7.8 by icepir8, ported by blight";

// GL WindowHandling Functions
BOOL
SDLGL_Init( void )
{
	const SDL_VideoInfo *videoInfo;
	Uint32 videoFlags = 0;
	SDL_Surface *tricon_surface;

	/* Initialize SDL */
	/*LOG( "(II) Initializing SDL..." );
	if ( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
		LOG( "(EE) Error initializing SDL: %s", SDL_GetError() );
		return FALSE;
	}*/

	/* Video Info */
	LOG( "(II) Getting video info..." );
	if ( !(videoInfo = SDL_GetVideoInfo()) ) {
		LOG( "(EE) Video query failed: %s", SDL_GetError() );
		SDL_Quit();
		return FALSE;
	}

	/* Setting the video mode */
	videoFlags |= SDL_OPENGL | SDL_GL_DOUBLEBUFFER | SDL_HWPALETTE;

	if ( videoInfo->hw_available )
		videoFlags |= SDL_HWSURFACE;
	else
		videoFlags |= SDL_SWSURFACE;

	if ( videoInfo->blit_hw )
		videoFlags |= SDL_HWACCEL;

	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	/*glEnable(	GL_POINT_SMOOTH );		glHint( GL_POINT_SMOOTH_HINT, window.flags.gleGL_POINT_SMOOTH_HINT );
	glEnable(	GL_LINE_SMOOTH );			glHint( GL_LINE_SMOOTH_HINT, window.flags.gleGL_LINE_SMOOTH_HINT );
	glEnable(	GL_POLYGON_SMOOTH );	glHint( GL_POLYGON_SMOOTH_HINT, window.flags.gleGL_POLYGON_SMOOTH_HINT );
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, window.flags.gleGL_PERSPECTIVE_CORRECTION_HINT );*/


 	/* tr icon */
/*	tricon_surface = SDL_CreateRGBSurfaceFrom( tricon_image.pixel_data, tricon_image.width, tricon_image.height, tricon_image.bytes_per_pixel,
																						 tricon_image.width * tricon_image.bytes_per_pixel, 0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF );
	if ( tricon_surface == NULL )
		LOG( "(WW) Creating SDL_Surface from tricon_image: %s", SDL_GetError() );
	else
		SDL_WM_SetIcon( tricon_surface, NULL );
*/
	LOG( "(II) Setting video mode %dx%dx%d...", window.width, window.height, window.ubpp );
	if ( !(window.screen = SDL_SetVideoMode( window.width, window.height, window.ubpp, videoFlags )) )
	{
		LOG( "(EE) Error setting videomode %dx%dx%d: %s", window.width, window.height, window.ubpp, SDL_GetError() );
		SDL_Quit();
		return FALSE;
	}

	SDL_WM_SetCaption( szAppName, "TR64GL" );

	SDLGL_InitState();

	return TRUE;
}

void SDLGL_UnInit(void)
{
	SDL_Quit();
}

BOOL SDLGL_SwapBuffers(void)
{	
	//glFlush();
	// Lothos: No No No No! its implied in SwapBuffers, and calling it 
	// manually is BAD!

	// FPS display
	static int framecount = 0, lastticks = 0;
	int nowticks = SDL_GetTicks();

	framecount++;

	if ( nowticks > (lastticks + 5000) )
	{
		char caption[4097];

		snprintf( caption, 4097, "%s - %.3f FPS", szAppName, (float)framecount / 5.0f );
		SDL_WM_SetCaption( caption, "TR64GL" );

		lastticks = nowticks;
		framecount = 0;
	}

	SDL_GL_SwapBuffers();
	return TRUE;
}

void
SDLGL_ToggleFullScreen( void )
{
	SDL_WM_ToggleFullScreen( window.screen );
}

#define MAX_CACHED_TEXTURES 4096 //512
GLuint tList[MAX_CACHED_TEXTURES];

void CreateDisplayLists();
void SDLGL_InitState(void)
{
	glViewport(0, 0, window.width, window.height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, window.width, window.height, 0, 0, 1);

	glGenTextures(MAX_CACHED_TEXTURES, tList);

	// position viewer
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0f);

	glShadeModel(GL_SMOOTH);

	glAlphaFunc(GL_GEQUAL,0.005f);
	glDisable(GL_ALPHA_TEST);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);

	glFrontFace(GL_CCW);
	glDisable(GL_CULL_FACE);

	glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);

//	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // config option

	glEnable(GL_BLEND);
	glEnable(GL_ALPHA_TEST);

	CreateDisplayLists();

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
}

/*****************************************************************************/
/* Copiled Displaylists for speedup                                          */
/*****************************************************************************/
#define kListEnd2D          1
#define kListBegin2D160   160
#define kListBegin2D320   320
#define kListBegin2D640   640

#define kListBegin2D800   800
#define kListBegin2D960   960
#define kListBegin2D1024 1024
#define kListBegin2D1200 1200

const float proj_matrix[2][4][4] =
{
    /* 320 * 240 */
        {
                { 0.0062500f, 0.0000000f, 0.0000000f, 0.0000000f},
                { 0.0000000f,-0.0083333f, 0.0000000f, 0.0000000f},
                { 0.0000000f, 0.0000000f,-1.0000000f, 0.0000000f},
                {-1.0000000f, 1.0000000f, 0.0000000f, 1.0000000f}
        },
    /* 640 * 480 */
        {

                { 0.0031250f, 0.0000000f, 0.0000000f, 0.0000000f},
                { 0.0000000f,-0.0041667f, 0.0000000f, 0.0000000f},
                { 0.0000000f, 0.0000000f,-1.0000000f, 0.0000000f},
                {-1.0000000f, 1.0000000f, 0.0000000f, 1.0000000f}
        }
};

void CreateDisplayLists()
{

    /* Gil: Some precompiled display lists that will make
            the emulation slighty faster. 
            I think we shold use more of these, if possible. */
    /* niki: gil - this is a fantastic idea!!! */

    /* This routine should be called when you init your opengl graphics */
       
    /* BEGIN_2D_320 */

        glNewList(kListBegin2D320, GL_COMPILE);
            glPushAttrib(GL_ENABLE_BIT);
        
 //           glDisable(GL_DEPTH_TEST);
            glDisable(GL_CULL_FACE);
            glDisable(GL_ALPHA_TEST);
            glEnable(GL_BLEND);

            glMatrixMode(GL_PROJECTION);
            glPushMatrix();
               //** I use precalculated matrixes for speed 
            glLoadMatrixf((GLfloat *)(&(proj_matrix[0][0][0])));
            glMatrixMode(GL_MODELVIEW);
            glPushMatrix();
            glLoadIdentity();
        glEndList();

   //** BEGIN_2D_640 
        glNewList(kListBegin2D640, GL_COMPILE);
            glPushAttrib(GL_ENABLE_BIT);

            glDisable(GL_CULL_FACE);
//            glDisable(GL_DEPTH_TEST);
            glDisable(GL_ALPHA_TEST);
            glEnable(GL_BLEND);

            glMatrixMode(GL_PROJECTION);
            glPushMatrix();
               //** I use precalculated matrixes for speed 
            glLoadMatrixf((GLfloat *)(&(proj_matrix[1][0][0])));
            glMatrixMode(GL_MODELVIEW);
            glPushMatrix();
            glLoadIdentity();
        glEndList();

   //** BEGIN_Lights 
        glNewList(kListBeginLights, GL_COMPILE);
            glPushAttrib(GL_ENABLE_BIT);

            glMatrixMode(GL_PROJECTION);
            glPushMatrix();
               //** I use precalculated matrixes for speed 
            glLoadIdentity();
            glMatrixMode(GL_MODELVIEW);
            glPushMatrix();
            glLoadIdentity();
        glEndList();

    /* END_2D */
        glNewList(kListEnd2D, GL_COMPILE);
            glPopMatrix();
            glMatrixMode(GL_PROJECTION);
            glPopMatrix();
            glMatrixMode(GL_MODELVIEW);
            glPopAttrib();
            glMatrixMode(GL_PROJECTION);
        glEndList();
}
