#ifdef _WIN32
#include <windows.h>
#else
#include "wintypes.h"
#endif
#include <GL/gl.h>
//#include <glext.h>
#include <GL/glu.h>
#include <stdio.h>
#include <stdlib.h>

#include "rdp_registers.h"
#include "texture.h"
#include "debug.h"

//** externals
extern int				*pInterruptMask;
extern unsigned char	*pRDRAM;
extern unsigned char	*pIDMEM;
extern unsigned char	*pVIREG;
extern HWND hGraphics;
extern int ucode_version;
extern _u8 m_tmem[0x1000];
extern _u32 pltmode;

//** texture - caching
//#define MaxTileSize (1024 * 512)
//#define MaxTileSize (128 * 256)
#define MaxTileSize (256 * 256)

unsigned int num_cached_texture = 1;
/* added by hacktarux : it's the crc hash table, to speed up the searching process */
text_info* crc_pointer[0x10000];

//** global variables

//#define TBUFFERSIZE  256*256*4
#define TBUFFERSIZE  512*512*4
//#define TBUFFERSIZE  1024*1024*4

static unsigned char t0bfr[TBUFFERSIZE];
static unsigned char t1bfr[TBUFFERSIZE];

//static unsigned char tbuffer0[TBUFFERSIZE];
static unsigned char tbuffer1[TBUFFERSIZE];
//static unsigned char tbuffer2[TBUFFERSIZE];

_u32 TexMode = 0;
_u32 TexColor = 0;
BOOL swapped = FALSE;

//** prototypes
static int TestPower2(int width, int height, int *nw, int *nh);
void UpdateTile(t_tile *tmpTile);
void DumpTile2Bmp(_u8 *bffr,_u32 width, _u32 height,t_tile *t_p);

static void NCopyTextureRGBA32(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureRGBA16(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureCI8   (t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureCI4   (t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureIA16  (t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureIA8   (t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureIA4   (t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureI16   (t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureI8    (t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
static void NCopyTextureI4    (t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey);
//static void NCopyTextureYUV   (_u8 *i_src, _u8 *i_dst, _u32 sx, _u32 ex, _u32 sy, _u32 ey);

static NCopyTexture lpNCopyTexture[20] =
	{
		NCopyTextureCI4,
		NCopyTextureCI8,
		NCopyTextureRGBA16,
		NCopyTextureRGBA32,

		NCopyTextureCI4,    //  unsupported YUV modes
		NCopyTextureCI8,
		NCopyTextureRGBA16,
		NCopyTextureRGBA32,

		NCopyTextureCI4,
		NCopyTextureCI8,
		NCopyTextureRGBA16,
		NCopyTextureRGBA32,

		NCopyTextureIA4,
		NCopyTextureIA8,
		NCopyTextureIA16,
		NCopyTextureIA4,

		NCopyTextureI4,
		NCopyTextureI8,
		NCopyTextureI16,
		NCopyTextureI4
};
/******************************************************************************\
*                                                                              *
*   helper routines									  						   *
*                                                                              *
\******************************************************************************/

void fastScaleImage(int srcwidth, int srcheight,unsigned char *SrcBuff, int dstwidth,int dstheight, unsigned char *DstBuff);
void fastRepeatScaleImage(int srcwidth, int srcheight,unsigned char *SrcBuff,int rptwidth, int rptheight, int dstwidth,int dstheight, unsigned char *DstBuff);

#ifdef _WIN32
//This function will not verify if values being passed to function are valid.
//format must be RGB8 or RGBA8 (GL_RGB, GL_RGB8, GL_RGBA, GL_RGBA8)
//dataAlignment must be 1
//Returns 1 if success. 0 for failure, -5 for not yet implemented
GLint gluScaleImage_asm386optimized(GLenum format, GLboolean isUseLinearOrPointSample,
		GLint dataAlignment,
		GLint widthin, GLint heightin,
		GLubyte *datain,
		GLint widthout, GLint heightout,
		GLubyte *dataout)
{
	long components;		//Bits per pixel (BPP)
	long i, j, tempo;
	float sx, sy;
	long ii, jj;
	__asm
	{
		mov		edx, format
		cmp		edx, GL_RGB
		je		Address_GL_RGB
		cmp		edx, GL_RGB8
		je		Address_GL_RGB
		cmp		edx, GL_RGBA
		je		Address_GL_RGBA
		cmp		edx, GL_RGBA8
		je		Address_GL_RGBA
	}
	return -5;
	__asm
	{
		Address_GL_RGB:
		mov		components, 3
		jmp		FormatExit
		Address_GL_RGBA:
		mov		components, 4
		//jmp		FormatExit
		FormatExit:
	}
	if(dataAlignment==1)
	{
	__asm
	{
		//Need some temporary long here.
		mov		ecx, widthin
		dec		ecx
		mov		i, ecx
		mov		edx, widthout
		cmp		edx, 1
		jle		JUMPPOINT_a_q22bx
		dec		edx
		mov		j, edx
		fild	i
		fild	j
		fdivp	st(1), st(0)	; st(0) contains sx
		fstp	sx				; store into sx
		jmp		JUMPPOINT_a_q22bxsup
		JUMPPOINT_a_q22bx:
		fild	i				; have the FPU translate integer to float
		fstp	sx				; store into sx
		JUMPPOINT_a_q22bxsup:
		mov		ecx, heightin
		dec		ecx
		mov		i, ecx
		mov		edx, heightout
		cmp		edx, 1
		jle		JUMPPOINT_b_yy789
		dec		edx
		mov		j, edx
		fild	i
		fild	j
		fdivp	st(1), st(0)	; st(0) contains sy
		fstp	sy				; store into sy
		jmp		JUMPPOINT_b_yy789sup
		JUMPPOINT_b_yy789:
		fild	i				; have the FPU translate integer to float
		fstp	sy				; store into sy
		JUMPPOINT_b_yy789sup:
	}
	if(isUseLinearOrPointSample==GL_FALSE)
	{
		//Gone use point sampling
		__asm
		{
			//ebx is dataout
			//ecx as i
			//edx as j
			mov		ebx, dataout
			cmp		components, 3
			jne		ThirtyTwoBits
			//This is the 24 bit version
			xor		ecx, ecx
			cmp		ecx, heightout
			jge		ExitLoops
			EnterLoop1:
			mov		i, ecx
			fild	i
			fmul	sy
			fistp	ii
			xor		edx, edx
			cmp		edx, widthout
			jge		ExitLoop1
			EnterLoop2:
			mov		j, edx
			fild	j
			fmul	sx
			fistp	jj
			mov		esi, ii
			imul	esi, widthin
			add		esi, jj
			imul	esi, 3
			add		esi, datain
			mov		edi, ecx		; ecx contains i
			imul	edi, widthout
			add		edi, j
			imul	edi, 3
			add		edi, ebx
			mov		al, [esi]		; copy red component
			mov		[edi], al
			mov		al, [esi+1]		; copy green component
			mov		[edi+1], al
			mov		al, [esi+2]		; copy blue component
			mov		[edi+2], al
			inc		edx
			cmp		edx, widthout
			jl		EnterLoop2
			ExitLoop1:
			inc		ecx
			cmp		ecx, heightout
			jl		EnterLoop1
			jmp		ExitLoops		; skip 32 bit section
			ThirtyTwoBits:
			//This is the 32 bit version
			xor		ecx, ecx
			cmp		ecx, heightout
			jge		ExitLoops
			EnterLoop1b:
			mov		i, ecx
			fild	i
			fmul	sy
			fistp	ii
			xor		edx, edx
			cmp		edx, widthout
			jge		ExitLoop1b
			EnterLoop2b:
			mov		j, edx
			fild	j
			fmul	sx
			fistp	jj
			mov		esi, ii
			imul	esi, widthin
			add		esi, jj
			imul	esi, 4
			add		esi, datain
			mov		edi, ecx		; ecx contains i
			imul	edi, widthout
			add		edi, j
			imul	edi, 4
			add		edi, ebx
			mov		al, [esi]		; copy red component
			mov		[edi], al
			mov		al, [esi+1]		; copy green component
			mov		[edi+1], al
			mov		al, [esi+2]		; copy blue component
			mov		[edi+2], al
			mov		al, [esi+3]		; copy alpha component
			mov		[edi+3], al
			inc		edx
			cmp		edx, widthout
			jl		EnterLoop2b
			ExitLoop1b:
			inc		ecx
			cmp		ecx, heightout
			jl		EnterLoop1b
			ExitLoops:
		}
		return 1;			//Success
	}
	else //if(isUseLinearOrPointSample==GL_TRUE)
	{
		short old_FPU_Flag, new_FPU_Flag;	//Must be 16 bit (word)
		__asm
		{
			fstcw	old_FPU_Flag	; store control word
			mov		ax, old_FPU_Flag
			and		ax, 0xF7FF		; clear bits 11 and 10
			or		ax, 0x0C00		; set them to 1
			mov		new_FPU_Flag, ax
			fldcw	new_FPU_Flag	; load control word
		}
		if((sx<1.0)&&(sy<1.0))
		{
			long i0, i1, j0, j1;
			float falpha, fbeta, foneminusalpha, foneminusbeta;//, s1, s2;
			float one=1.0;
			unsigned char *src00, *src01, *src10, *src11;
			__asm
			{
				//ecx is i
				//edx is j
				cmp		components, 3
				jne		ThirtyTwoBits_66_78
				//24 bit section
				fld		one				; need this for later (1.0-falpha and 1.0 fbeta)
				xor		ecx, ecx
				cmp		ecx, heightout
				jge		ExitSuperly
				EnterLoop1_dashj:
				mov		i, ecx
				fild	i
				fmul	sy				; i0=long(float(i)*sy);
				fist	i0				; keep the value in FPU register for later 
				mov		eax, i0
				inc		eax
				mov		ebx, heightin
				cmp		eax, ebx
				jl		SkipIfStatement_firett_a
				dec		ebx
				mov		i1, ebx
				jmp		SkipIfStatement_firett_b
				SkipIfStatement_firett_a:
				mov		i1, eax
				SkipIfStatement_firett_b:
				fisub	i0					; this is falpha in st(0) - falpha=float(i)*sy-float(i0);
				fst		falpha
				fsub	st(0), st(1)		; this is foneminusalpha in st(0)
				fchs						; must change the sign
				fstp	foneminusalpha
				xor		edx, edx
				cmp		edx, widthout
				jge		ExitLoop2_sippy
				EnterLoop2_sippy:
				mov		j, edx
				fild	j
				fmul	sx				; float(j)*sx
				fist	j0				; keep this in FPU register for later
				mov		eax, j0
				inc		eax
				mov		ebx, widthin
				cmp		eax, ebx
				jl		SkipIfStatement_firett_c
				dec		ebx
				mov		j1, ebx
				jmp		SkipIfStatement_firett_d
				SkipIfStatement_firett_c:
				mov		j1, eax
				SkipIfStatement_firett_d:
				fisub	j0
				fst		fbeta
				fsub	st(0), st(1)		; this is foneminusbeta in st(0)
				fchs						; must change the sign
				fstp	foneminusbeta
				//Now esi is src00
				//Now edi is src01
				//Now ebx is src10
				//Now ecx is src11
				mov		eax, datain			; temporary (not enough registers to make this algorithm efficient)
				mov		esi, i0
				mov		edi, esi			; make a copy of i0 for efficiency
				imul	esi, widthin
				add		esi, j0
				imul	esi, 3
				add		esi, eax
				imul	edi, widthin
				add		edi, j1
				imul	edi, 3
				add		edi, eax
				mov		ebx, i1
				mov		ecx, ebx			; make a copy if i1 for efficiency
				imul	ebx, widthin
				add		ebx, j0
				imul	ebx, 3
				add		ebx, eax
				imul	ecx, widthin
				add		ecx, j1
				imul	ecx, 3
				add		ecx, eax
				mov		eax, i				; dst=dataout+(i*widthout+j)*3;
				imul	eax, widthout
				add		eax, edx			; edx is j in this case
				imul	eax, 3
				add		eax, dataout
				mov		edx, eax			; now we have address in dst or in this case, edx
				mov		src00, esi
				mov		src01, edi
				mov		src10, ebx
				mov		src11, ecx
				xor		eax, eax
				mov		al, [esi]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src00[0]*foneminusbeta
				mov		al, [edi]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src01[0]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	foneminusalpha		; *foneminusalpha
				mov		al, [ebx]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src10[0]*foneminusbeta
				mov		al, [ecx]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src11[0]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	falpha				; *falpha
				faddp	st(1), st(0)		; finally, add both parts to finish
				fistp	tempo				; store result in 32 bit integer
				mov		eax, tempo
				mov		byte ptr[edx], al	; dst[0] contains result now
				xor		eax, eax
				mov		al, [esi+1]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src00[1]*foneminusbeta
				mov		al, [edi+1]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src01[1]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	foneminusalpha		; *foneminusalpha
				mov		al, [ebx+1]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src10[1]*foneminusbeta
				mov		al, [ecx+1]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src11[1]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	falpha				; *falpha
				faddp	st(1), st(0)		; finally, add both parts to finish
				fistp	tempo				; store result in 32 bit integer
				mov		eax, tempo
				mov		byte ptr[edx+1], al	; dst[1] contains result now
				xor		eax, eax
				mov		al, [esi+2]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src00[2]*foneminusbeta
				mov		al, [edi+2]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src01[2]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	foneminusalpha		; *foneminusalpha
				mov		al, [ebx+2]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src10[2]*foneminusbeta
				mov		al, [ecx+2]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src11[2]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	falpha				; *falpha
				faddp	st(1), st(0)		; finally, add both parts to finish
				fistp	tempo				; store result in 32 bit integer
				mov		eax, tempo
				mov		byte ptr[edx+2], al	; dst[2] contains result now
				//finit							; Clear the FPU stack!
				mov		edx, j
				inc		edx
				cmp		edx, widthout
				jl		EnterLoop2_sippy
				ExitLoop2_sippy:
				mov		ecx, i
				inc		ecx
				cmp		ecx, heightout
				jl		EnterLoop1_dashj
				jmp		ExitSuperly
				//32 bit section
				ThirtyTwoBits_66_78:
				fld		one				; need this for later (1.0-falpha and 1.0 fbeta)
				xor		ecx, ecx
				cmp		ecx, heightout
				jge		ExitSuperly
				EnterLoop1_dashj_b:
				mov		i, ecx
				fild	i
				fld		sy
				fmulp	st(1), st(0)
				fist	i0				; keep the value in FPU register for later
				mov		eax, i0
				inc		eax
				mov		ebx, heightin
				cmp		eax, ebx
				jl		SkipIfStatement_firett_a_b
				dec		ebx
				mov		i1, ebx
				jmp		SkipIfStatement_firett_b_b
				SkipIfStatement_firett_a_b:
				mov		i1, eax
				SkipIfStatement_firett_b_b:
				fild	i0
				fsubp	st(1), st(0)		; this is falpha in st(0)
				fst		falpha
				fsub	st(0), st(1)		; this is foneminusalpha in st(0)
				fchs						; must change the sign
				fstp	foneminusalpha
				xor		edx, edx
				cmp		edx, widthout
				jge		ExitLoop2_sippy_b
				EnterLoop2_sippy_b:
				mov		j, edx
				fild	j
				fld		sx
				fmulp	st(1), st(0)
				fist	j0				; keep this in FPU register for later
				mov		eax, j0
				inc		eax
				mov		ebx, widthin
				cmp		eax, ebx
				jl		SkipIfStatement_firett_c_b
				dec		ebx
				mov		j1, ebx
				jmp		SkipIfStatement_firett_d_b
				SkipIfStatement_firett_c_b:
				mov		j1, eax
				SkipIfStatement_firett_d_b:
				fld		j0
				fsubp	st(1), st(0)		; this is fbeta in st(0)
				fst		fbeta
				fsub	st(0), st(1)		; this is foneminusbeta in st(0)
				fchs						; must change the sign
				fstp	foneminusbeta
				//Now esi is src00
				//Now edi is src01
				//Now ebx is src10
				//Now ecx is src11
				mov		eax, datain			; temporary (not enough registers to make this algorithm efficient)
				mov		esi, i0
				mov		edi, esi			; make a copy of i0 for efficiency
				imul	esi, widthin
				add		esi, j0
				imul	esi, 4				; 4 bytes per pixel
				add		esi, eax
				imul	edi, widthin
				add		edi, j1
				imul	edi, 4				; 4 bytes per pixel
				add		edi, eax
				mov		ebx, i1
				mov		ecx, ebx			; make a copy if i1 for efficiency
				imul	ebx, widthin
				add		ebx, j0
				imul	ebx, 4				; 4 bytes per pixel
				add		ebx, eax
				imul	ecx, widthin
				add		ecx, j1
				imul	ecx, 4				; 4 bytes per pixel
				add		ecx, eax
				mov		eax, i				; dst=dataout+(i*widthout+j)*3;
				imul	eax, widthout
				add		eax, edx			; edx is j
				imul	eax, 4				; 4 bytes per pixel
				add		eax, dataout
				mov		edx, eax			; now we have address in dst
				xor		eax, eax
				mov		al, [esi]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src00[0]*foneminusbeta
				mov		al, [edi]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src01[0]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	foneminusalpha		; *foneminusalpha
				mov		al, [ebx]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src10[0]*foneminusbeta
				mov		al, [ecx]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src11[0]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	falpha				; *falpha
				faddp	st(1), st(0)		; finally, add both parts to finish
				fistp	tempo				; store result in 32 bit integer
				mov		eax, tempo
				mov		byte ptr[edx], al	; dst[0] contains result now
				mov		al, [esi+1]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src00[1]*foneminusbeta
				mov		al, [edi+1]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src01[1]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	foneminusalpha		; *foneminusalpha
				mov		al, [ebx+1]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src10[1]*foneminusbeta
				mov		al, [ecx+1]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src11[1]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	falpha				; *falpha
				faddp	st(1), st(0)		; finally, add both parts to finish
				fistp	tempo				; store result in 32 bit integer
				mov		eax, tempo
				mov		byte ptr[edx+1], al	; dst[1] contains result now
				mov		byte ptr[edx+1], al	; dst[1] contains result now*/
				mov		al, [esi+2]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src00[2]*foneminusbeta
				mov		al, [edi+2]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src01[2]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	foneminusalpha		; *foneminusalpha
				mov		al, [ebx+2]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src10[2]*foneminusbeta
				mov		al, [ecx+2]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src11[2]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	falpha				; *falpha
				faddp	st(1), st(0)		; finally, add both parts to finish
				fistp	tempo				; store result in 32 bit integer
				mov		eax, tempo
				mov		byte ptr[edx+2], al	; dst[2] contains result now
				mov		al, [esi+3]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src00[3]*foneminusbeta
				mov		al, [edi+3]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src01[3]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	foneminusalpha		; *foneminusalpha
				mov		al, [ebx+3]
				mov		tempo, eax
				fild	tempo
				fmul	foneminusbeta		; src10[3]*foneminusbeta
				mov		al, [ecx+3]
				mov		tempo, eax
				fild	tempo
				fmul	fbeta				; src11[3]*fbeta
				faddp	st(1), st(0)		; add them together
				fmul	falpha				; *falpha
				faddp	st(1), st(0)		; finally, add both parts to finish
				fistp	tempo				; store result in 32 bit integer
				mov		eax, tempo
				mov		byte ptr[edx+3], al	; dst[3] contains result now
				mov		edx, j
				inc		edx
				cmp		edx, widthout
				jl		EnterLoop2_sippy_b
				ExitLoop2_sippy_b:
				mov		ecx, i
				inc		ecx
				cmp		ecx, heightout
				jl		EnterLoop1_dashj_b
				ExitSuperly:
				finit			; a single pop would suffice here
			}
		}
		else
		{
			long i0, i1;
			long j0, j1;
			long sum;
			__asm
			{
				//ecx is i
				//edx is j
				cmp		components, 3
				jne		ThirtyTwoBits_xcd99
				//24 bit section
				xor		ecx, ecx
				cmp		ecx, heightout
				jge		ExitLoop1_a_stuzz
				EnterLoop1_a_stuzz:
				mov		i, ecx
				fild	i
				fmul	sy
				fistp	i0			; store as integer
				mov		eax, i0
				inc		eax
				mov		ebx, heightin
				cmp		eax, ebx
				jl		Skipper_tyui
				dec		ebx
				mov		i1, ebx
				jmp		Skipper_tyui2
				Skipper_tyui:
				mov		i1, eax
				Skipper_tyui2:
				xor		edx, edx
				cmp		edx, widthout
				jge		ExitLoop2_c178
				Enter_Loop2_c178:
				mov		j, edx
				fild	j
				fmul	sx
				fistp	j0
				mov		eax, j0
				inc		eax
				mov		ebx, widthin
				cmp		eax, ebx
				jl		Skipper_tyui_zz
				dec		ebx
				mov		j1, ebx
				jmp		Skipper_tyui_zz2
				Skipper_tyui_zz:
				mov		j1, eax
				Skipper_tyui_zz2:
				mov		edi, i
				imul	edi, widthout
				add		edi, j
				imul	edi, 3
				add		edi, dataout
				//Compute average of pixels in the rectangle (i0,j0)-(i1,j1)
				xor		ebx, ebx		; need for computing sum
				mov		sum, 0			; sum=0
				mov		ecx, i0
				cmp		ecx, i1
				jg		Exit_rtt1
				Enter_rtt1:
				mov		edx, j0
				cmp		edx, j1
				jg		Exit_rtt2
				Enter_rtt2:
				mov		esi, ecx
				imul	esi, widthin
				add		esi, edx
				imul	esi, 3
				add		esi, datain
				mov		bl, [esi]				; sum+=long(*(datain+(ii*widthin+jj)*3));
				add		sum, ebx
				inc		edx
				cmp		edx, j1
				jle		Enter_rtt2
				Exit_rtt2:
				inc		ecx
				cmp		ecx, i1
				jle		Enter_rtt1
				Exit_rtt1:
				mov		eax, sum		; sum/=(j1-j0+1)*(i1-i0+1);
				mov		ebx, j1
				sub		ebx, j0
				inc		ebx
				mov		ecx, i1
				sub		ecx, i0
				inc		ecx
				imul	ecx, ebx
				cdq
				idiv	ecx				; result is in eax
				mov		[edi], al		; dst[0]=(unsigned char)sum; (lower part of AX register)
				///////////////////////////REPEAT the above
				xor		ebx, ebx		; need for computing sum
				mov		sum, 0			; sum=0
				mov		ecx, i0
				cmp		ecx, i1
				jg		Exit_rtt1_b
				Enter_rtt1_b:
				mov		edx, j0
				cmp		edx, j1
				jg		Exit_rtt2_b
				Enter_rtt2_b:
				mov		esi, ecx
				imul	esi, widthin
				add		esi, edx
				imul	esi, 3
				inc		esi
				add		esi, datain
				mov		bl, [esi]				; sum+=long(*(datain+(ii*widthin+jj)*3+1));
				add		sum, ebx
				inc		edx
				cmp		edx, j1
				jle		Enter_rtt2_b
				Exit_rtt2_b:
				inc		ecx
				cmp		ecx, i1
				jle		Enter_rtt1_b
				Exit_rtt1_b:
				mov		eax, sum		; sum/=(j1-j0+1)*(i1-i0+1);
				mov		ebx, j1
				sub		ebx, j0
				inc		ebx
				mov		ecx, i1
				sub		ecx, i0
				inc		ecx
				imul	ecx, ebx
				cdq
				idiv	ecx				; result is in eax
				mov		[edi+1], al		; dst[1]=(unsigned char)sum; (lower part of AX register)
				///////////////////////////REPEAT the above
				xor		ebx, ebx		; need for computing sum
				mov		sum, 0			; sum=0
				mov		ecx, i0
				cmp		ecx, i1
				jg		Exit_rtt1_c
				Enter_rtt1_c:
				mov		edx, j0
				cmp		edx, j1
				jg		Exit_rtt2_c
				Enter_rtt2_c:
				mov		esi, ecx
				imul	esi, widthin
				add		esi, edx
				imul	esi, 3
				add		esi, 2
				add		esi, datain
				mov		bl, [esi]				; sum+=long(*(datain+(ii*widthin+jj)*3+2));
				add		sum, ebx
				inc		edx
				cmp		edx, j1
				jle		Enter_rtt2_c
				Exit_rtt2_c:
				inc		ecx
				cmp		ecx, i1
				jle		Enter_rtt1_c
				Exit_rtt1_c:
				mov		eax, sum		; sum/=(j1-j0+1)*(i1-i0+1);
				mov		ebx, j1
				sub		ebx, j0
				inc		ebx
				mov		ecx, i1
				sub		ecx, i0
				inc		ecx
				imul	ecx, ebx
				cdq
				idiv	ecx				; result is in eax
				mov		[edi+2], al		; dst[2]=(unsigned char)sum; (lower part of AX register)
				mov		edx, j
				inc		edx
				cmp		edx, widthout
				jl		Enter_Loop2_c178
				ExitLoop2_c178:
				mov		ecx, i
				inc		ecx
				cmp		ecx, heightout
				jl		EnterLoop1_a_stuzz
				ExitLoop1_a_stuzz:
				jmp		CLEAR_IF_1023
				//32 bit section
				ThirtyTwoBits_xcd99:
				xor		ecx, ecx
				cmp		ecx, heightout
				jge		ExitLoop1_a_stuzz_32
				EnterLoop1_a_stuzz_32:
				mov		i, ecx
				fild	i
				fmul	sy
				fistp	i0			; store as integer
				mov		eax, i0
				inc		eax
				mov		ebx, heightin
				cmp		eax, ebx
				jl		Skipper_tyui_32
				dec		ebx
				mov		i1, ebx
				jmp		Skipper_tyui2_32
				Skipper_tyui_32:
				mov		i1, eax
				Skipper_tyui2_32:
				xor		edx, edx
				cmp		edx, widthout
				jge		ExitLoop2_c178_32
				Enter_Loop2_c178_32:
				mov		j, edx
				fild	j
				fmul	sx
				fistp	j0
				mov		eax, j0
				inc		eax
				mov		ebx, widthin
				cmp		eax, ebx
				jl		Skipper_tyui_zz_32
				dec		ebx
				mov		j1, ebx
				jmp		Skipper_tyui_zz2_32
				Skipper_tyui_zz_32:
				mov		j1, eax
				Skipper_tyui_zz2_32:
				mov		edi, i
				imul	edi, widthout
				add		edi, j
				imul	edi, 4
				add		edi, dataout
				//Compute average of pixels in the rectangle (i0,j0)-(i1,j1)
				xor		ebx, ebx		; need for computing sum
				mov		sum, 0			; sum=0
				mov		ecx, i0
				cmp		ecx, i1
				jg		Exit_rtt1_32
				Enter_rtt1_32:
				mov		edx, j0
				cmp		edx, j1
				jg		Exit_rtt2_32
				Enter_rtt2_32:
				mov		esi, ecx
				imul	esi, widthin
				add		esi, edx
				imul	esi, 4
				add		esi, datain
				mov		bl, [esi]				; sum+=long(*(datain+(ii*widthin+jj)*3));
				add		sum, ebx
				inc		edx
				cmp		edx, j1
				jle		Enter_rtt2_32
				Exit_rtt2_32:
				inc		ecx
				cmp		ecx, i1
				jle		Enter_rtt1_32
				Exit_rtt1_32:
				mov		eax, sum		; sum/=(j1-j0+1)*(i1-i0+1);
				mov		ebx, j1
				sub		ebx, j0
				inc		ebx
				mov		ecx, i1
				sub		ecx, i0
				inc		ecx
				imul	ecx, ebx
				cdq
				idiv	ecx				; result is in eax
				mov		[edi], al		; dst[0]=(unsigned char)sum; (lower part of AX register)
				///////////////////////////REPEAT the above
				xor		ebx, ebx		; need for computing sum
				mov		sum, 0			; sum=0
				mov		ecx, i0
				cmp		ecx, i1
				jg		Exit_rtt1_b_32
				Enter_rtt1_b_32:
				mov		edx, j0
				cmp		edx, j1
				jg		Exit_rtt2_b_32
				Enter_rtt2_b_32:
				mov		esi, ecx
				imul	esi, widthin
				add		esi, edx
				imul	esi, 4
				inc		esi
				add		esi, datain
				mov		bl, [esi]				; sum+=long(*(datain+(ii*widthin+jj)*3+1));
				add		sum, ebx
				inc		edx
				cmp		edx, j1
				jle		Enter_rtt2_b_32
				Exit_rtt2_b_32:
				inc		ecx
				cmp		ecx, i1
				jle		Enter_rtt1_b_32
				Exit_rtt1_b_32:
				mov		eax, sum		; sum/=(j1-j0+1)*(i1-i0+1);
				mov		ebx, j1
				sub		ebx, j0
				inc		ebx
				mov		ecx, i1
				sub		ecx, i0
				inc		ecx
				imul	ecx, ebx
				cdq
				idiv	ecx				; result is in eax
				mov		[edi+1], al		; dst[1]=(unsigned char)sum; (lower part of AX register)
				///////////////////////////REPEAT the above
				xor		ebx, ebx		; need for computing sum
				mov		sum, 0			; sum=0
				mov		ecx, i0
				cmp		ecx, i1
				jg		Exit_rtt1_c_32
				Enter_rtt1_c_32:
				mov		edx, j0
				cmp		edx, j1
				jg		Exit_rtt2_c_32
				Enter_rtt2_c_32:
				mov		esi, ecx
				imul	esi, widthin
				add		esi, edx
				imul	esi, 3
				add		esi, 2
				add		esi, datain
				mov		bl, [esi]				; sum+=long(*(datain+(ii*widthin+jj)*3+2));
				add		sum, ebx
				inc		edx
				cmp		edx, j1
				jle		Enter_rtt2_c_32
				Exit_rtt2_c_32:
				inc		ecx
				cmp		ecx, i1
				jle		Enter_rtt1_c_32
				Exit_rtt1_c_32:
				mov		eax, sum		; sum/=(j1-j0+1)*(i1-i0+1);
				mov		ebx, j1
				sub		ebx, j0
				inc		ebx
				mov		ecx, i1
				sub		ecx, i0
				inc		ecx
				imul	ecx, ebx
				cdq
				idiv	ecx				; result is in eax
				mov		[edi+2], al		; dst[2]=(unsigned char)sum; (lower part of AX register)
				///////////////////////////REPEAT the above
				xor		ebx, ebx		; need for computing sum
				mov		sum, 0			; sum=0
				mov		ecx, i0
				cmp		ecx, i1
				jg		Exit_rtt1_d_32
				Enter_rtt1_d_32:
				mov		edx, j0
				cmp		edx, j1
				jg		Exit_rtt2_d_32
				Enter_rtt2_d_32:
				mov		esi, ecx
				imul	esi, widthin
				add		esi, edx
				imul	esi, 3
				add		esi, 3
				add		esi, datain
				mov		bl, [esi]				; sum+=long(*(datain+(ii*widthin+jj)*3+2));
				add		sum, ebx
				inc		edx
				cmp		edx, j1
				jle		Enter_rtt2_d_32
				Exit_rtt2_d_32:
				inc		ecx
				cmp		ecx, i1
				jle		Enter_rtt1_d_32
				Exit_rtt1_d_32:
				mov		eax, sum		; sum/=(j1-j0+1)*(i1-i0+1);
				mov		ebx, j1
				sub		ebx, j0
				inc		ebx
				mov		ecx, i1
				sub		ecx, i0
				inc		ecx
				imul	ecx, ebx
				cdq
				idiv	ecx				; result is in eax
				mov		[edi+3], al		; dst[3]=(unsigned char)sum; (lower part of AX register)
				mov		edx, j
				inc		edx
				cmp		edx, widthout
				jl		Enter_Loop2_c178_32
				ExitLoop2_c178_32:
				mov		ecx, i
				inc		ecx
				cmp		ecx, heightout
				jl		EnterLoop1_a_stuzz_32
				ExitLoop1_a_stuzz_32:
				CLEAR_IF_1023:
			}
		}
		__asm
		{
			fldcw	old_FPU_Flag	; load control word
		}
	}
	return 1;		//Success
	}	//dataAlignment==1

	return -5;		//Other data alignments are not implemented
}
#endif // _WIN32
_u32 *Palette8;
_u32 Palette8IA[256];
_u32 Palette8RGBA[256];
int tstart;
int todd;

void DelTexture() // modified by hacktarux to free the hash table
{
	unsigned int i, *v;
	v = malloc(num_cached_texture*sizeof(int));
	for (i=1; i < num_cached_texture; i ++)
		v[i-1] = i;
	glDeleteTextures(num_cached_texture-1,v);
	free(v);
	num_cached_texture=1;
	for (i=0; i<0x10000; i++)
	{
		while (crc_pointer[i])
		{
			text_info *aux = crc_pointer[i]->next;
			free(crc_pointer[i]);
			crc_pointer[i] = aux;
		}
	}
}

static int TestPower2(int width, int height, int *nw, int *nh)
{
	int w;
	int temp, NewWidth = *nw, NewHeight = *nh;
	
//** test if width has the power of 2
	temp = 1;
	for (w = 0; w <= 9;w++)
	{
		NewWidth = temp;
		temp = temp << 1;
		if (NewWidth >= width)
		{
			break;
		}
	}
//** test if height has the power of 2
	temp = 1;
	for (w = 0; w <= 9;w++)
	{
		NewHeight = temp;
		temp = temp << 1;
		if (NewHeight >= height)
		{
			break;
		}
	}
	if (NewWidth > 512) NewWidth = 512;
	if (NewHeight > 512) NewHeight = 512;

//** return
	if ((NewWidth == width) && (NewHeight == height)) return 1;
	else
	{
		*nw = NewWidth;
		*nh = NewHeight;
		return 0;
	}
	return 1;
}

/******************************************************************************\
*                                                                              *
*   Texture Handling								  						   *
*                                                                              *
\******************************************************************************/
_u32 txwidth;
extern int TextWidth;

_u32 CRCTile(t_tile *tmpTile)
{
	//_u8 *src = (_u8*)(tmpTile->addr);
//	_u32 *src = (_u32*)(tmpTile->addr);
	_u32 *src = (_u32*)&pRDRAM[(tmpTile->offset)];

	_s32  tuls, tlrs, tult, tlrt;
	_u32 palette = 0;

	_u32 i;
	_u32 j;

    int width  = tmpTile->Width;//  tmpTile->lrs - tmpTile->uls;
    int height = tmpTile->Height; //tmpTile->lrt - tmpTile->ult;
	_u32 tsize;
	_u32 tpitch;

	_u8 size   = tmpTile->size;
	_u8 format = tmpTile->format;
	_u32 CRC = 0;

	CRC ^= tmpTile->offset;
//	CRC ^= (((_u32)tmpTile->uls << 8) | ((_u32)tmpTile->ult << 3));
//	CRC ^= (_u32)tmpTile->lrs << 11 ^ (_u32)tmpTile->lrt << 6;
	CRC ^= (((_u32)tmpTile->mirrors << 13) | ((_u32)tmpTile->mirrort << 19));
	CRC ^= (((_u32)tmpTile->clamps << 14) | ((_u32)tmpTile->clampt << 20));
//	CRC ^= (_u32)tmpTile->offset;

	if (src == 0) return -1;

	if (format == 2)
	{
		switch (size)
		{
		case 0:
			palette = PaletteCRC[tmpTile->palette];
//			tmpTile->texture = tmpTile->ttexture[tmpTile->palette];
//			tmpTile->CRC = tmpTile->tCRC[tmpTile->palette];
			break;
		case 1:
			palette = PaletteCRC[0];
			break;
		default:
			palette = 0;
			break;
		}
	}
	
	CRC ^= palette;

	tuls = tmpTile->uls;
	tlrs = tmpTile->lrs;
	tult = tmpTile->ult;
	tlrt = tmpTile->lrt;
	
	//tsize = (tlrs - tuls+1)*(tlrt - tult+1);

	tsize = tmpTile->Width * tmpTile->Height;

	if (tsize > MaxTileSize) {tmpTile->CRC = -2; return -2;}
	if (tsize < 0)  {tmpTile->CRC = -2; return -2;}
	if (tmpTile->Width < 0)  {tmpTile->CRC = -3; return -3;}
	
	tsize = tmpTile->Width;
	tpitch = tmpTile->Pitch;
	switch (tmpTile->size)
	{
	case 0:
		tsize >>= 3;
		tpitch >>= 3;
		break;
	case 1:
		tsize >>= 2;
		tpitch >>= 2;
		break;
	case 2:
		tsize >>= 1;
		tpitch >>= 1;
		break;
	}

	if (tmpTile->LineWidth)
		tpitch = tmpTile->LineWidth;
	else
		tpitch = tmpTile->line;
	tpitch /= 4;

	if (tpitch == 0)
		tpitch = 1;

	if (tsize == 0)
		tsize = 1;

	for (i = 0; i < tmpTile->Height;i+=1)
		for (j = 0; j < tsize;j+=1)
		{
			int tIdx = i * tpitch + j;
			CRC = (CRC>>(7 + (i & 0x0f))) | (CRC<<(25 - (i & 0x0f)));
			CRC += ((_u32*)src)[tIdx];
			CRC = (CRC>>(23 - (j & 0x0f))) | (CRC<<(9 + (j & 0x0f)));
			CRC ^= ((_u32*)src)[tIdx];
		}

	return CRC;
}

int DecompressTile(t_tile *tmpTile)
{
	_u8 *src  = (_u8*)(tmpTile->addr);
	_u8 *dest;

	_s32  tuls, tlrs, tult, tlrt;
	_u32 palette = 0;

    int width  = tmpTile->lrs - tmpTile->uls;
    int height = tmpTile->lrt - tmpTile->ult;
	_u32 tsize;

	_u8 size   = tmpTile->size;
	_u8 format = tmpTile->format;
	_u32 CRC = 0;

	//tmpTile->texture = tmpTile->ttexture[0];
	//tmpTile->CRC = tmpTile->tCRC[0];

	pltmode = (rdp_reg.mode_h & 0x0000c000);

	swapped = tmpTile->WdSwpd;

	if (src == 0) return -1;
/*
	if (format == 2)
	{
		switch (size)
		{
		case 0:
			palette = 0;
//			palette = PaletteCRC[tmpTile->palette];
//			tmpTile->texture = tmpTile->ttexture[tmpTile->palette];
//			tmpTile->CRC = tmpTile->tCRC[tmpTile->palette];
			break;
		case 1:
			palette = PaletteCRC[0];
			break;
		default:
			palette = 0;
			break;
		}
	}
*/	
	tuls = tmpTile->uls;
	tlrs = tmpTile->lrs;
	tult = tmpTile->ult;
	tlrt = tmpTile->lrt;

//	tstart = tmpTile->uls & 3;
	tstart = tmpTile->offset;
	todd = 0;//tmpTile->ult & 1;
	
	tsize = tmpTile->Width * tmpTile->Height;

	if (tsize > MaxTileSize) {tmpTile->CRC = -2; return -2;}

	CRC = tmpTile->lCRC;  //CRCTile(tmpTile);

	if (CRC == tmpTile->CRC)
		return CRC;

	width = tmpTile->Width;
	
	dest = tmpTile->texture;

	txwidth = rdp_reg.TextureImage.width;
/*
	if ((rdp_reg.TextureImage.size == (tmpTile->size + 1)) && (tmpTile->size == 0))
		txwidth += txwidth;
*/
	if (width < 0) width = - width;
	if (height < 0) height = - height;

	height++;

	tuls = tmpTile->uls;
	tlrs = tmpTile->lrs;
	tult = tmpTile->ult;
	tlrt = tmpTile->lrt;

	if (tmpTile->LineWidth)
		txwidth = tmpTile->LineWidth;
	else
		txwidth = tmpTile->line;

	lpNCopyTexture[format << 2 | size](tmpTile, src, dest, 0, tmpTile->Width - 1, 0, tmpTile->Height);

#ifdef DUMP_TILE
		DumpTile2Bmp(dest, tmpTile->Width, tmpTile->Height,tmpTile);
#endif
	CRC = tmpTile->CRC;
	tmpTile->lCRC = CRC;
	return CRC;
}

void loadtile(int tile)
{
	t_tile *t_p = &rdp_reg.td[rdp_reg.tile + tile];
	t_p->Number = rdp_reg.tile + tile;
//	UpdateTile(t_p);
	DecompressTile(t_p);
}

_u32 loadCRC(int tile)
{
	t_tile *t_p = &rdp_reg.td[rdp_reg.tile + tile];
	UpdateTile(t_p);
	t_p->CRC = CRCTile(t_p);
	return t_p->CRC;
}

int ModTexture(int tTexMode,int Length)
{
	int i; 
	_u32 tmp = 0;
	_u32 *tile0 = (_u32*)&t0bfr[0];
	_u32 *tile1 = (_u32*)&t1bfr[0];
	_u32 *dest = (_u32*)&tbuffer1[0];

//	return 1;

	for (i=0;i<Length;i++) dest[i]=0xffffffff;
/**/
	switch (tTexMode & TEX_MASK)
	{
	case Tex0:
		{
			_u32 *tClr = (_u32*)tile0;
			_u32 *dClr = (_u32*)dest;

			for (i=0;i<Length;i++)
				*dClr++ = (*tClr++ & 0xffffff) | 0xff000000;
		}
		break;
	case Tex1:
		{
			_u32 *tClr = (_u32*)tile1;
			_u32 *dClr = (_u32*)dest;

			for (i=0;i<Length;i++)
				*dClr++ = (*tClr++ & 0xffffff) | 0xff000000;
		}
		break;
	case InvT0:
		{
			_u32 *tClr = (_u32*)tile0;
			_u32 *dClr = (_u32*)dest;

			for (i=0;i<Length;i++)
				*dClr++ = ((0xffffff ^ *tClr++) & 0xffffff) | 0xff000000;
		}
		break;
	case InvT1:
		{
			_u32 *tClr = (_u32*)tile1;
			_u32 *dClr = (_u32*)dest;

			for (i=0;i<Length;i++)
				*dClr++ = ((0xffffff ^ *tClr++) & 0xffffff) | 0xff000000;
		}
		break;
	case T0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				_u8 Alpha = t0Clr[3];
				dClr[0] = Alpha;
				dClr[1] = Alpha;
				dClr[2] = Alpha;
				dClr += 4;
				t0Clr += 4;
			}
		}
		break;
	case T1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				_u8 Alpha = t1Clr[3];
				dClr[0] = Alpha;
				dClr[1] = Alpha;
				dClr[2] = Alpha;
				dClr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				_u8 Alpha = t0Clr[3] ^ 0xff;
				dClr[0] = Alpha;
				dClr[1] = Alpha;
				dClr[2] = Alpha;
				dClr += 4;
				t0Clr += 4;
			}
		}
		break;
	case InvT1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				_u8 Alpha = t1Clr[3] ^ 0xff;
				dClr[0] = Alpha;
				dClr[1] = Alpha;
				dClr[2] = Alpha;
				dClr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex0 | Tex1:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t0Clr[0]) * (_u32)t1Clr[0])>>8);
				dClr[1] = (_u8)(((1 + (_u32)t0Clr[1]) * (_u32)t1Clr[1])>>8);
				dClr[2] = (_u8)(((1 + (_u32)t0Clr[2]) * (_u32)t1Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex0 | InvT0:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t0Clr[0]) * (_u32)t0Clr[0])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t0Clr[1]) * (_u32)t0Clr[1])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t0Clr[2]) * (_u32)t0Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex0 | InvT1:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t1Clr[0]) * (_u32)t0Clr[0])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t1Clr[1]) * (_u32)t0Clr[1])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t1Clr[2]) * (_u32)t0Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT0 | Tex1:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t0Clr[0]) * (_u32)t1Clr[0])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t0Clr[1]) * (_u32)t1Clr[1])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t0Clr[2]) * (_u32)t1Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT1 | Tex1:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t1Clr[0]) * (_u32)t1Clr[0])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t1Clr[1]) * (_u32)t1Clr[1])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t1Clr[2]) * (_u32)t1Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT0 | InvT1:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t0Clr[0]) * (_u32)t1Clr[0])>>8) ^ 0x0ff;
				dClr[1] = (_u8)(((1 + (_u32)t0Clr[1]) * (_u32)t1Clr[1])>>8) ^ 0x0ff;
				dClr[2] = (_u8)(((1 + (_u32)t0Clr[2]) * (_u32)t1Clr[2])>>8) ^ 0x0ff;
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex0 | T0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t0Clr[0]) * (_u32)t0Clr[3])>>8);
				dClr[1] = (_u8)(((1 + (_u32)t0Clr[1]) * (_u32)t0Clr[3])>>8);
				dClr[2] = (_u8)(((1 + (_u32)t0Clr[2]) * (_u32)t0Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex0 | InvT0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t0Clr[3]) * (_u32)t0Clr[0])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t0Clr[3]) * (_u32)t0Clr[1])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t0Clr[3]) * (_u32)t0Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT0 | T0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t0Clr[0]) * (_u32)t0Clr[3])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t0Clr[1]) * (_u32)t0Clr[3])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t0Clr[2]) * (_u32)t0Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT0 | InvT0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t0Clr[0]) * (_u32)t0Clr[3])>>8) ^ 0x0ff;
				dClr[1] = (_u8)(((1 + (_u32)t0Clr[1]) * (_u32)t0Clr[3])>>8) ^ 0x0ff;
				dClr[2] = (_u8)(((1 + (_u32)t0Clr[2]) * (_u32)t0Clr[3])>>8) ^ 0x0ff;
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex0 | T1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t0Clr[0]) * (_u32)t1Clr[3])>>8);
				dClr[1] = (_u8)(((1 + (_u32)t0Clr[1]) * (_u32)t1Clr[3])>>8);
				dClr[2] = (_u8)(((1 + (_u32)t0Clr[2]) * (_u32)t1Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex0 | InvT1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t1Clr[3]) * (_u32)t0Clr[0])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t1Clr[3]) * (_u32)t0Clr[1])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t1Clr[3]) * (_u32)t0Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT0 | T1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t0Clr[0]) * (_u32)t1Clr[3])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t0Clr[1]) * (_u32)t1Clr[3])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t0Clr[2]) * (_u32)t1Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT0 | InvT1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t0Clr[0]) * (_u32)t1Clr[3])>>8) ^ 0x0ff;
				dClr[1] = (_u8)(((1 + (_u32)t0Clr[1]) * (_u32)t1Clr[3])>>8) ^ 0x0ff;
				dClr[2] = (_u8)(((1 + (_u32)t0Clr[2]) * (_u32)t1Clr[3])>>8) ^ 0x0ff;
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex1 | T0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t1Clr[0]) * (_u32)t0Clr[3])>>8);
				dClr[1] = (_u8)(((1 + (_u32)t1Clr[1]) * (_u32)t0Clr[3])>>8);
				dClr[2] = (_u8)(((1 + (_u32)t1Clr[2]) * (_u32)t0Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex1 | InvT0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t0Clr[3]) * (_u32)t1Clr[0])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t0Clr[3]) * (_u32)t1Clr[1])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t0Clr[3]) * (_u32)t1Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT1 | T0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t1Clr[0]) * (_u32)t0Clr[3])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t1Clr[1]) * (_u32)t0Clr[3])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t1Clr[2]) * (_u32)t0Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT1 | InvT0MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t1Clr[0]) * (_u32)t0Clr[3])>>8) ^ 0x0ff;
				dClr[1] = (_u8)(((1 + (_u32)t1Clr[1]) * (_u32)t0Clr[3])>>8) ^ 0x0ff;
				dClr[2] = (_u8)(((1 + (_u32)t1Clr[2]) * (_u32)t0Clr[3])>>8) ^ 0x0ff;
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex1 | T1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t1Clr[0]) * (_u32)t1Clr[3])>>8);
				dClr[1] = (_u8)(((1 + (_u32)t1Clr[1]) * (_u32)t1Clr[3])>>8);
				dClr[2] = (_u8)(((1 + (_u32)t1Clr[2]) * (_u32)t1Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case Tex1 | InvT1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t1Clr[3]) * (_u32)t1Clr[0])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t1Clr[3]) * (_u32)t1Clr[1])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t1Clr[3]) * (_u32)t1Clr[2])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT1 | T1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((256 - (_u32)t1Clr[0]) * (_u32)t1Clr[3])>>8);
				dClr[1] = (_u8)(((256 - (_u32)t1Clr[1]) * (_u32)t1Clr[3])>>8);
				dClr[2] = (_u8)(((256 - (_u32)t1Clr[2]) * (_u32)t1Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT1 | InvT1MultAlpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[0] = (_u8)(((1 + (_u32)t1Clr[0]) * (_u32)t1Clr[3])>>8) ^ 0x0ff;
				dClr[1] = (_u8)(((1 + (_u32)t1Clr[1]) * (_u32)t1Clr[3])>>8) ^ 0x0ff;
				dClr[2] = (_u8)(((1 + (_u32)t1Clr[2]) * (_u32)t1Clr[3])>>8) ^ 0x0ff;
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	default:
		{
			int i = 0;
		}
		break;
	}
/**//*
	if (tTexMode & Tex0)
	{
		_u8 *tClr = (_u8*)tile0;
		_u8 *dClr = (_u8*)dest;
		_u32 tR,tG,tB;

		for (i=0;i<Length;i++)
		{
			tR = (_u32)tClr[0] + 1;
			tG = (_u32)tClr[1] + 1;
			tB = (_u32)tClr[2] + 1;
			dClr[0] = (_u8)((((_u32)dClr[0] * tR)>>8) & 0xff);
			dClr[1] = (_u8)((((_u32)dClr[1] * tG)>>8) & 0xff);
			dClr[2] = (_u8)((((_u32)dClr[2] * tB)>>8) & 0xff);
			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & Tex1)
	{
		_u8 *tClr = (_u8*)tile1;
		_u8 *dClr = (_u8*)dest;
		_u32 tR,tG,tB;

		for (i=0;i<Length;i++)
		{
			tR = (_u32)tClr[0] + 1;
			tG = (_u32)tClr[1] + 1;
			tB = (_u32)tClr[2] + 1;
			dClr[0] = (_u8)((((_u32)dClr[0] * tR)>>8) & 0xff);
			dClr[1] = (_u8)((((_u32)dClr[1] * tG)>>8) & 0xff);
			dClr[2] = (_u8)((((_u32)dClr[2] * tB)>>8) & 0xff);
			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & InvT0)
	{
		_u8 *tClr = (_u8*)tile0;
		_u8 *dClr = (_u8*)dest;
		_u32 tR,tG,tB;

		for (i=0;i<Length;i++)
		{
			tR = 256 - (_u32)tClr[0];
			tG = 256 - (_u32)tClr[1];
			tB = 256 - (_u32)tClr[2];
			dClr[0] = (_u8)((((_u32)dClr[0] * tR)>>8) & 0xff);
			dClr[1] = (_u8)((((_u32)dClr[1] * tG)>>8) & 0xff);
			dClr[2] = (_u8)((((_u32)dClr[2] * tB)>>8) & 0xff);
			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & InvT1)
	{
		_u8 *tClr = (_u8*)tile1;
		_u8 *dClr = (_u8*)dest;
		_u32 tR,tG,tB;

		for (i=0;i<Length;i++)
		{
			tR = 256 - (_u32)tClr[0];
			tG = 256 - (_u32)tClr[1];
			tB = 256 - (_u32)tClr[2];
			dClr[0] = (_u8)((((_u32)dClr[0] * tR)>>8) & 0xff);
			dClr[1] = (_u8)((((_u32)dClr[1] * tG)>>8) & 0xff);
			dClr[2] = (_u8)((((_u32)dClr[2] * tB)>>8) & 0xff);
			dClr += 4;
			tClr += 4;
		}
	}
	if (tTexMode & T0MultAlpha)
	{
		_u8 *tClr = (_u8*)tile0;
		_u8 *dClr = (_u8*)dest;
		_u32 alpha;

		for (i=0;i<Length;i++)
		{
			alpha = (_u32)tClr[3] + 1;

			dClr[0] = (_u8)((((_u32)dClr[0] * alpha)>>8) & 0xff);
			dClr[1] = (_u8)((((_u32)dClr[1] * alpha)>>8) & 0xff);
			dClr[2] = (_u8)((((_u32)dClr[2] * alpha)>>8) & 0xff);

			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & T1MultAlpha)
	{
		_u8 *tClr = (_u8*)tile1;
		_u8 *dClr = (_u8*)dest;
		_u32 alpha;

		for (i=0;i<Length;i++)
		{
			alpha = (_u32)tClr[3] + 1;

			dClr[0] = (_u8)((((_u32)dClr[0] * alpha)>>8) & 0xff);
			dClr[1] = (_u8)((((_u32)dClr[1] * alpha)>>8) & 0xff);
			dClr[2] = (_u8)((((_u32)dClr[2] * alpha)>>8) & 0xff);

			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & InvT0MultAlpha)
	{
		_u8 *tClr = (_u8*)tile0;
		_u8 *dClr = (_u8*)dest;
		_u32 alpha;

		for (i=0;i<Length;i++)
		{
			alpha = 256 - (_u32)tClr[3];

			dClr[0] = (_u8)((((_u32)dClr[0] * alpha)>>8) & 0xff);
			dClr[1] = (_u8)((((_u32)dClr[1] * alpha)>>8) & 0xff);
			dClr[2] = (_u8)((((_u32)dClr[2] * alpha)>>8) & 0xff);

			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & InvT1MultAlpha)
	{
		_u8 *tClr = (_u8*)tile1;
		_u8 *dClr = (_u8*)dest;
		_u32 alpha;

		for (i=0;i<Length;i++)
		{
			alpha = 256 - (_u32)tClr[3];

			dClr[0] = (_u8)((((_u32)dClr[0] * alpha)>>8) & 0xff);
			dClr[1] = (_u8)((((_u32)dClr[1] * alpha)>>8) & 0xff);
			dClr[2] = (_u8)((((_u32)dClr[2] * alpha)>>8) & 0xff);

			dClr += 4;
			tClr += 4;
		}
	}
*/	
/**/
	switch (tTexMode & TEX_ALPHA_MASK)
	{
	case AlphaT0:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = t0Clr[3];
				dClr += 4;
				t0Clr += 4;
			}
		}
		break;
	case AlphaT1:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = t1Clr[3];
				dClr += 4;
				t1Clr += 4;
			}
		}
		break;
	case InvT0Alpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = t0Clr[3] ^ 0xff;
				dClr += 4;
				t0Clr += 4;
			}
		}
		break;
	case InvT1Alpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = t1Clr[3] ^ 0xff;
				dClr += 4;
				t1Clr += 4;
			}
		}
		break;
	case AlphaT0 | AlphaT1:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = (_u8)(((1 + (_u32)t0Clr[3]) * (_u32)t1Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case AlphaT0 | InvT0Alpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = (_u8)(((256 - (_u32)t0Clr[3]) * (_u32)t0Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case AlphaT0 | InvT1Alpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = (_u8)(((256 - (_u32)t1Clr[3]) * (_u32)t0Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case AlphaT1 | InvT0Alpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = (_u8)(((256 - (_u32)t0Clr[3]) * (_u32)t1Clr[3])>>8);
				dClr += 4;
				t0Clr += 4;
				t1Clr += 4;
			}
		}
		break;
	case AlphaT1 | InvT1Alpha:
		{
			_u8 *t0Clr = (_u8*)tile0;
			_u8 *t1Clr = (_u8*)tile1;
			_u8 *dClr = (_u8*)dest;

			for (i=0;i<Length;i++)
			{
				dClr[3] = (_u8)(((1 + (_u32)t1Clr[3]) * (_u32)t1Clr[3])>>8) ^ 0x0ff;
				dClr += 4;
				t1Clr += 4;
			}
		}
		break;
	default:
		{
			int i = 0;
		}
		break;
	}
/**/ /*
	if (tTexMode & AlphaT0)
	{
		_u8 *tClr = (_u8*)tile0;
		_u8 *dClr = (_u8*)dest;
		_u32 alpha;

		for (i=0;i<Length;i++)
		{
			alpha = (_u32)tClr[3] + 1;

			dClr[3] = (_u8)((((_u32)dClr[3] * alpha)>>8) & 0xff);

			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & AlphaT1)
	{
		_u8 *tClr = (_u8*)tile1;
		_u8 *dClr = (_u8*)dest;
		_u32 alpha;

		for (i=0;i<Length;i++)
		{
			alpha = (_u32)tClr[3] + 1;

			dClr[3] = (_u8)((((_u32)dClr[3] * alpha)>>8) & 0xff);

			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & InvT0Alpha)
	{
		_u8 *tClr = (_u8*)tile0;
		_u8 *dClr = (_u8*)dest;
		_u32 alpha;

		for (i=0;i<Length;i++)
		{
			alpha = 256 - (_u32)tClr[3];

			dClr[3] = (_u8)((((_u32)dClr[3] * alpha)>>8) & 0xff);

			dClr += 4;
			tClr += 4;
		}
	}

	if (tTexMode & InvT1Alpha)
	{
		_u8 *tClr = (_u8*)tile1;
		_u8 *dClr = (_u8*)dest;
		_u32 alpha;

		for (i=0;i<Length;i++)
		{
			alpha = 256 - (_u32)tClr[3];

			dClr[3] = (_u8)((((_u32)dClr[3] * alpha)>>8) & 0xff);

			dClr += 4;
			tClr += 4;
		}
	}
*/
	return 1;
}

void TexMirrorS(int width,int height,_u8 *pBuff)
{
	_u32 *d = (_u32*)&tbuffer1[0];
	_u32 *s = (_u32*)&pBuff[0];
	int x,y;
	for (y=0;y<height;y++)
	{
		s = &((_u32*)pBuff)[y * width];
		for (x=0;x<width;x++)
		{
			*d++ = *s++;
		}
		s--;
		for (x=0;x<width;x++)
		{
			*d++ = *s--;
		}
	}
	d = (_u32*)&pBuff[0];
	s = (_u32*)&tbuffer1[0];
	for (x = 0; x < (width * height); x++)
	{
		*d++ = *s++;
		*d++ = *s++;
	}
}

void TexMirrorT(int width,int height,_u8 *pBuff)
{
	_u32 *d = &((_u32*)pBuff)[width * height];
	_u32 *s = &((_u32*)pBuff)[width * height - 1];
	int x,y;
	for (y=0;y<height;y++)
	{
		for (x=0;x<width;x++)
		{
			*d++ = *s--;
		}
	}
}

void TexClampRepeatS(int rcnt,int width,int height,_u8 *pBuff)
{
	_u32 *d = (_u32*)&tbuffer1[0];
	_u32 *s = (_u32*)&pBuff[0];

	int x,y,r;

	for (x = 0; x < (width * height); x++)
	{
		*d++ = *s++;
	}

	s = (_u32*)&tbuffer1[0];
	d = (_u32*)&pBuff[0];

	for (y=0;y<height;y++)
	{
		s = &((_u32*)tbuffer1)[y * width];

		for (x=0;x<width;x++)
		{
			*d++ = *s++;
		}

		s--; // repeat last position

		for(r=1;r<rcnt-1;r++)
		{
			for (x=0;x<width;x++)
			{
				*d++ = *s;
			}
		}
		s = &((_u32*)tbuffer1)[y * width];
		for (x=0;x<width;x++)
		{
			*d++ = *s;
		}
	}
}

void TexRepeatS(int rcnt,int width,int height,_u8 *pBuff)
{
	_u32 *d = (_u32*)&tbuffer1[0];
	_u32 *s = (_u32*)&pBuff[0];
	int x,y,r;

	for (x = 0; x < (width * height); x++)
	{
		*d++ = *s++;
	}

	d = (_u32*)&pBuff[0];
	s = (_u32*)&tbuffer1[0];

	for (y=0;y<height;y++)
	{
		for(r=0;r<rcnt;r++)
		{
			s = &((_u32*)tbuffer1)[y * width];
			for (x=0;x<width;x++)
			{
				*d++ = *s++;
			}
		}
	}
}

void TexClampRepeatT(int rcnt,int width,int height,_u8 *pBuff)
{
	_u32 *d = &((_u32*)pBuff)[width * height];
	_u32 *s;// = &((_u32*)pBuff)[0];
	_u32 *s2 = &((_u32*)pBuff)[width * (height-1)];
	int x,y,r;
/*
	s = &((_u32*)pBuff)[0];
	for (y=0;y<height;y++)
	{
		s2 = s;
		for (x=0;x<width;x++)
		{
			*d++ = *s++;
			*d++ = *s++;
		}
	}
*/
	for (r=1;r<rcnt-1;r++)
	{
		for (y=0;y<height;y++)
		{
			s = s2;
			for (x=0;x<width;x++)
			{
				*d++ = *s++;
			}
		}
	}
	for (y=0;y<height;y++)
	{
		s = &((_u32*)pBuff)[0];
		for (x=0;x<width;x++)
		{
			*d++ = *s++;
		}
	}
}

void TexRepeatT(int rcnt,int width,int height,_u8 *pBuff)
{
	_u32 *d = &((_u32*)pBuff)[width * height];
	_u32 *s;// = &((_u32*)pBuff)[0];
	int x,y,r;

	for (r=1;r<rcnt;r++)
	{
		s = &((_u32*)pBuff)[0];
		for (y=0;y<height;y++)
		{
			for (x=0;x<width;x++)
			{
				*d++ = *s++;
			}
		}
	}
}

int bmpcnt = 0;
byte bmpHdr[] = { 0x42, 0x4D, 0x36, 0x84,
                  0x03, 0x00, 0x00, 0x00,
				  0x00, 0x00, 0x36, 0x00,
				  0x00, 0x00, 0x28, 0x00,
				  0x00, 0x00, 0x40, 0x01,
				  0x00, 0x00, 0xF0, 0x00,
				  0x00, 0x00, 0x01, 0x00,
				  0x18, 0x00, 0x00, 0x00,
				  0x00, 0x00, 0x00, 0x84,
				  0x03, 0x00, 0xC4, 0x0E,
				  0x00, 0x00, 0xC4, 0x0E,
				  0x00, 0x00, 0x00, 0x00,
				  0x00, 0x00, 0x00, 0x00,
				  0x00, 0x00};

void DumpTile2Bmp(_u8 *bffr,_u32 width, _u32 height,t_tile *t_p)
{
	//t_tile *t_p = &rdp_reg.td[0];
	t_tile *lt_p = &rdp_reg.td[7];

	int pidx = t_p->palette;

    static char *format[]   = { "RGBA", "YUV", "CI", "IA", "I", "?", "?", "?" };
    static char *size[]     = { "4bit", "8bit", "16bit", "32bit" };
    static char *cm[]       = { "NOMIRROR/WARP(NOCLAMP)", "MIRROR", "CLAMP", "MIRROR&CLAMP" };

	int x,y;
	_u32 *tbffr = (_u32*)bffr;
	char bmpFileName[128] = "";
	char sMsg[256] = "";

	FILE *bmpFile;
	byte tline[1024 * 4];

	*(_u32*)&bmpHdr[2] = (_u32) ((width * height * 3) + 0x036);

	*(_u32*)&bmpHdr[0x12] = width;
	*(_u32*)&bmpHdr[0x16] = height;

	*(_u32*)&bmpHdr[0x22] = (_u32) ((width * height * 3));

	sprintf(bmpFileName,"Tiles\\Tile%04i_RGB.bmp",bmpcnt,t_p->CRC);
	bmpFile = fopen(bmpFileName,"wb");

	if (t_p->size) pidx = 16;

    fwrite(bmpHdr, 0x36, 1, bmpFile);
	
	for ( y = (int)height - 1;y >= 0; y--)
	{
		tbffr = (_u32*)bffr;
		tbffr += (y * width);
		for (x = 0; x < (int)width; x++)
		{
			tline[x*3+0] = ((_u8*)tbffr)[2];
			tline[x*3+1] = ((_u8*)tbffr)[1];
			tline[x*3+2] = ((_u8*)tbffr)[0];
			tbffr++;
		}

		fwrite(tline, ((width * 3) + 3) & 0x0fffc, 1, bmpFile);
	}
	fclose(bmpFile);

	sprintf(bmpFileName,"Tiles\\Tile%04i_Alpha.bmp",bmpcnt,t_p->CRC);
	bmpFile = fopen(bmpFileName,"wb");

	if (t_p->size) pidx = 16;

    fwrite(bmpHdr, 0x36, 1, bmpFile);
	
	tbffr = (_u32*)bffr;
	
	for ( y = (int)height - 1;y >= 0; y--)
	{
		tbffr = (_u32*)bffr;
		tbffr += (y * width);
		for (x = 0; x < (int)width; x++)
		{
			tline[x*3+0] = ((_u8*)tbffr)[3];
			tline[x*3+1] = ((_u8*)tbffr)[3];
			tline[x*3+2] = ((_u8*)tbffr)[3];
			tbffr++;
		}

		fwrite(tline, ((width * 3) + 3) & 0x0fffc, 1, bmpFile);
	}
	fclose(bmpFile);

	sprintf(bmpFileName,"Tiles\\Tile%04i_Info.txt",bmpcnt++,t_p->CRC);
	bmpFile = fopen(bmpFileName,"wb");

	sprintf(sMsg,"Tile = %i\r\n",t_p->Number);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"Width = %i, Height = %i\r\n",width,height);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"format = %s, size = %s, line = %i, tmem = %08x, Palette = %i\r\n",
	format[t_p->format],size[t_p->size],t_p->line,t_p->tmem,t_p->palette);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"clamp_s = %x, mirror_s = %x, mask_s = %x, shift_s = %x\r\n",
	t_p->clamps,t_p->mirrors,t_p->masks,t_p->shifts);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"clamp_t = %x, mirror_t = %x, mask_t = %x, shift_t = %x\r\n",
	t_p->clampt,t_p->mirrort,t_p->maskt,t_p->shiftt);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"uls = %x, ult = %x, lrs = %x, lrt = %x\r\n",
	t_p->uls,t_p->ult,t_p->lrs,t_p->lrt);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"Swaped = %x, Pitch = %x\r\n, LineWidth = %x\r\n",
	t_p->WdSwpd,t_p->Pitch,t_p->LineWidth);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"CRC = %08x, Palette CRC = %08x\r\n",
	t_p->CRC,PaletteCRC[pidx]);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"TexMode = %08x\r\n",
	TexMode);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"Address = %08x, Offset = %08x\r\n",
	t_p->addr,t_p->offset);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

	sprintf(sMsg,"load_tile:\r\nuls = %x, ult = %x, lrs = %x, lrt = %x\r\n",
	lt_p->uls,lt_p->ult,lt_p->lrs,lt_p->lrt);

	fwrite(sMsg, strlen(sMsg), 1, bmpFile);

/*
	t_p->Width,t_p->Height,
	t_p->Texture_on,
	t_p->SScale,t_p->TScale,

	t_p->set_by,
	t_p->dxt,
	t_p->sts_uls,t_p->sts_ult,t_p->sts_lrs,t_p->sts_lrt,
*/
	fclose(bmpFile);
}

/* hacktarux : we need to initialize the hash table before starting the emulation */
void init_cache()
{
	int i;
	for (i=0; i<0x10000; i++)
		crc_pointer[i] = NULL;
}

static t_tile tsa_p;

extern int aTilew;
extern int aTileh;
extern float Multw;
extern float Multh;

t_tile tCurr;
t_tile *tC_p = &tCurr;

int SetActiveTexture()
{
	_u32 new_width, new_height;
	text_info *p;

	t_tile *t_p = &rdp_reg.td[rdp_reg.tile];

	t_tile *t0_p = &rdp_reg.td[rdp_reg.tile];
	t_tile *t1_p = &rdp_reg.td[rdp_reg.tile + 1];

	int offset;
	int Size;

	_u32 CRC = 0;
	_u32 CRC1 = 0;
	_u32 CRC_lu = 0;

	_u64 addr = rdp_reg.TextureImage.addr;

	int  size = (int)t_p->size;
	int  format = (int)t_p->format;
	_u32  palette = 0;
	int twidth = t_p->Width;
	int width = rdp_reg.TextureImage.width;

	int realwidth = t_p->Width;
	int realheight = t_p->Height;

    int stwidth  = t_p->Width;
    int stheight = t_p->Height;

	int t0width;
	int t0height;
	int t1width;
	int t1height;

	_s32 t0w,t0h,t1w,t1h;
	int mt0w,mt0h,mt1w,mt1h;

	int ScaleAdjW = 0;
	int ScaleAdjH = 0;

	int MaxW = 0;
	int MaxH = 0;

	int FixUpS = 0;
	int FixUpT = 0;

	if (PaletteCRC[16] != ~InvPaletteCRC[16])
		UpdatePalette();

	Multw = 1.0f;
	Multh = 1.0f;

	tC_p->clamps = 1;
	tC_p->clampt = 1;

	if (((TexMode & (LoadTile0)) != 0) && ((TexMode & (LoadTile1)) != 0))
	{
		FixUpS = t0_p->clamps ^ t1_p->clamps;
		FixUpT = t0_p->clampt ^ t1_p->clampt;
	}

	if (TexMode & (LoadTile0))
	{
		CRC = (loadCRC(0));// ^(
//			(t0_p->clamps << 4) |
//			(t0_p->clampt << 3) |
//			(t0_p->mirrors << 2) |
//			(t0_p->mirrort << 1)));

		tC_p->clamps &= t0_p->clamps;
		tC_p->clampt &= t0_p->clampt;
	}

	if (TexMode & (LoadTile1))
	{
		CRC1 = (loadCRC(1));// ^(
//			(t1_p->clamps << 4) |
//			(t1_p->clampt << 3) |
//			(t1_p->mirrors << 2) |
//			(t1_p->mirrort << 1)));
		CRC1 = (CRC1 << 5) | (CRC1 >> 27);

		tC_p->clamps &= t1_p->clamps;
		tC_p->clampt &= t1_p->clampt;
	}

	CRC ^= CRC1;

	if (stwidth < 0) stwidth = -stwidth;
	if (stheight < 0) stheight = -stheight;

	if ((rdp_reg.TextureImage.size == (t_p->size + 1)) && (width > 1))
		width += width;

	if (((_s32)stwidth * (_s32)stheight) > ((_s32)TBUFFERSIZE - 0x100)/4)
	{
		return 0;
	}

	Size = (stwidth * stheight);

	palette = 0;
	twidth = stwidth;

	if (width <= 1)
		width = twidth;

	offset =  ((t_p->ult * width) + t_p->uls);// 8-bit

	/* is texture already cached ? */
	
	CRC_lu = (CRC ^ (CRC>>16))&0xFFFF;

	if (crc_pointer[CRC_lu])
	{
		text_info *p = crc_pointer[CRC_lu];
		while (p)
		{
			if ((p->info.TexID == CRC)
				//&& (p->info.Size == Size)
				//&& (p->info.offset == offset)
				&& (p->info.TexMode == TexMode)
				//&& (p->info.TexColor == TexColor)
				&& (p->info.Palette == palette)
				)
			{ 
				glBindTexture(GL_TEXTURE_2D, p->info.NextID);
				aTilew = p->info.aTilew;
				aTileh = p->info.aTileh;
				Multw = p->info.Multw;
				Multh = p->info.Multh;
				return 1;
			}
			p = p->next;
		}
	}

//** texture is not cached yet... so we cache it

	aTilew = 0;
	aTileh = 0;
	Multw = 1.0f;
	Multh = 1.0f;

	if (TexMode & (LoadTile0))
		loadtile(0);
	if (TexMode & (LoadTile1))
		loadtile(1);

	LOG_TO_FILE("\t\tActive Texture CRC = %08x",CRC);

	/* here, we store the texture info of the newly cached texture */
	{
		p = crc_pointer[CRC_lu];
		if (!p)
		{
			p = malloc(sizeof(text_info));
			p->next = NULL;
			crc_pointer[CRC_lu] = p;
		}
		else
		{
			while(p->next) p = p->next;
			p->next = malloc(sizeof(text_info));
			p = p->next;
			p->next = NULL;
		}
		p->info.TexID		= CRC;
		p->info.offset		= offset;
		p->info.TexMode		= TexMode;
		p->info.TexColor	= TexColor;
		p->info.Palette		= palette;
		p->info.Size		= Size;
		p->info.NextID		= num_cached_texture;
	}
	glBindTexture(GL_TEXTURE_2D, num_cached_texture);

	if ((TexMode & (LoadTile1)) && !(TexMode & (LoadTile0)))
	{
		aTilew = 1;
		aTileh = 1;
	}

	{
		t0w = 128;
		t0h = 128;
		t1w = 128;
		t1h = 128;

		mt0w = 1;
		mt0h = 1;
		mt1w = 1;
		mt1h = 1;

		if (TexMode & (LoadTile1)) 
		{
			if (TestPower2(t1_p->Width, t1_p->Height, &t1width, &t1height))
			{
				t1width = t1_p->Width;
				t1height = t1_p->Height;
			}

			t1w = t1_p->Width;
			t1h = t1_p->Height;

			if(t1_p->mirrors)
			{
				t1w *= 2;
			}

			if(t1_p->mirrort)
			{
				t1h *= 2;
			}

			if(t1_p->shifts<11)
			{
				t1w *= 1 << t1_p->shifts;
				ScaleAdjW += t1_p->shifts;
			}
			else
			{
				t1w /= 1 << (16 - t1_p->shifts);
				ScaleAdjW += t1_p->shifts - 16;
			}

			if(t1_p->shiftt<11)
			{
				t1h *= 1 << t1_p->shiftt;
				ScaleAdjH += t1_p->shiftt;
			}
			else
			{
				t1h /= 1 << (16 - t1_p->shiftt);
				ScaleAdjH += t1_p->shiftt - 16;
			}
		}

		if (TexMode & (LoadTile0)) 
		{
			if (TestPower2(t0_p->Width, t0_p->Height , &t0width, &t0height))
			{
				t0width = t0_p->Width;
				t0height = t0_p->Height;
			}

			t0w = t0_p->Width;
			t0h = t0_p->Height;

			if(t0_p->mirrors)
			{
				t0w *= 2;
			}

			if(t0_p->mirrort)
			{
				t0h *= 2;
			}

			if(t0_p->shifts<11)
			{
				t0w *= 1 << t0_p->shifts;
				ScaleAdjW -= t0_p->shifts;
			}
			else
			{
				t0w /= 1 << (16 - t0_p->shifts);
				ScaleAdjW -= t0_p->shifts - 16;
			}

			if(t0_p->shiftt<11)
			{
				t0h *= 1 << t0_p->shiftt;
				ScaleAdjH -= t0_p->shiftt;
			}
			else
			{
				t0h /= 1 << (16 - t0_p->shiftt);
				ScaleAdjH -= t0_p->shiftt - 16;
			}
		}

		if (!(TexMode & (LoadTile0))) 
		{
			t0w = t1w;
			t0h = t1h;
			t0width = t1width;
			t0height = t1height ;
			ScaleAdjW = 0;
			ScaleAdjH = 0;
		}

		if (!(TexMode & (LoadTile1))) 
		{
			t1w = t0w;
			t1h = t0h;
			t1width = t0width;
			t1height = t0height ;
			ScaleAdjW = 0;
			ScaleAdjH = 0;
		}

		if (t0w == 0)
			t0w++;
		if (t0h == 0)
			t0h++;
		if (t1w == 0)
			t1w++;
		if (t1h == 0)
			t1h++;

		if (t0w != t1w)
		{
			if (t0w > t1w)
				mt1w *= (t0w / t1w);
			else
				mt0w *= (t1w / t0w);
			Multw = mt0w;
		}

		if (t0h != t1h)
		{
			if (t0h > t1h)
				mt1h *= (t0h / t1h);
			else
				mt0h *= (t1h / t0h);
			Multh = mt0h;
		}

		Multw = mt0w;
		Multh = mt0h;

		if(FixUpS)
		{
			mt0w <<= 2;
			mt1w <<= 2;
			Multw *= 4;
		}
		
		if(FixUpT)
		{
			mt0h <<= 2;
			mt1h <<= 2;
			Multh *= 4;
		}

		if (mt0w == 0)
			mt0w++;
		if (mt0h == 0)
			mt0h++;
		if (mt1w == 0)
			mt1w++;
		if (mt1h == 0)
			mt1h++;
	}

	ScaleAdjW = -ScaleAdjW;
	ScaleAdjH = -ScaleAdjH;

	{
		int lt0w,lt1w;
		int lt0h,lt1h;

		lt0w = t0_p->Width;
		lt1w = t1_p->Width;
		lt0h = t0_p->Height;
		lt1h = t1_p->Height;

		if (TexMode & (LoadTile0))
		{
			if (t0_p->mirrors) lt0w <<= 1;
			if (t0_p->mirrort) lt0h <<= 1;
			if (ScaleAdjW > 0) lt0w <<= ScaleAdjW;
			if (ScaleAdjH > 0) lt0h <<= ScaleAdjH;
		}

		if (TexMode & (LoadTile1))
		{
			if (t1_p->mirrors) lt1w <<= 1;
			if (t1_p->mirrort) lt1h <<= 1;
			if (ScaleAdjW < 0) lt1w <<= -ScaleAdjW;
			if (ScaleAdjH < 0) lt1h <<= -ScaleAdjH;
		}

		if(FixUpS)
		{
			lt0w <<= 2;
			lt1w <<= 2;
		}
		
		if(FixUpT)
		{
			lt0h <<= 2;
			lt1h <<= 2;
		}

		if (!(TexMode & (LoadTile0)))
		{
			lt0w = lt1w;
			lt0h = lt1h;
		}

		if (!(TexMode & (LoadTile1)))
		{
			lt1w = lt0w;
			lt1h = lt0h;
		}

		MaxW = (lt0w > lt1w)? lt0w:lt1w;
		MaxH = (lt0h > lt1h)? lt0h:lt1h;
	}

	if (TexMode & (LoadTile0)) 
	{
		_u32 *tsrc = (_u32*)t0_p->texture;
		_u32 *tdst = (_u32*)t0bfr;

		_u32 tw = t0width;
		_u32 th = t0height;

		if (ScaleAdjW > 0)
			tw <<= ScaleAdjW;
		if (ScaleAdjH > 0)
			th <<= ScaleAdjH;

		if (MaxW > 512)
			tw = (tw * 512) / MaxW;
		if (MaxH > 512)
			th = (th * 512) / MaxH;

//		fastScaleImage(t0_p->Width, t0_p->Height,(unsigned char*)tsrc,
//							tw, th, (unsigned char*)tdst);

//		gluScaleImage_asm386optimized(GL_RGBA, GL_FALSE, 1,t0_p->Width, t0_p->Height,(unsigned char*)tsrc,
//							tw, th,(unsigned char*)tdst);
		gluScaleImage(GL_RGBA,t0_p->Width, t0_p->Height,GL_UNSIGNED_BYTE,(unsigned char*)tsrc,
							tw, th,GL_UNSIGNED_BYTE, (unsigned char*)tdst);

		if (t0_p->mirrors)
		{
			TexMirrorS(tw, th,t0bfr);
			tw *= 2;
		}

		if (t0_p->mirrort)
		{
			TexMirrorT(tw, th,t0bfr);
			th *= 2;
		}

		if (mt0w > 1)
		{
			if(t0_p->clamps)
				TexClampRepeatS(mt0w,tw, th,t0bfr);
			else
				TexRepeatS(mt0w,tw, th,t0bfr);
			tw *= mt0w;
		}

		if (mt0h > 1)
		{
			if(t0_p->clampt)
				TexClampRepeatT(mt0h,tw, th,t0bfr);
			else
				TexRepeatT(mt0h,tw, th,t0bfr);
			th *= mt0h;
		}

		t0width = tw;
		t0height = th;
	}

	if (TexMode & (LoadTile1)) 
	{
		_u32 *tsrc = (_u32*)t1_p->texture;
		_u32 *tdst = (_u32*)t1bfr;

		_u32 tw = t1width;
		_u32 th = t1height;

		if (ScaleAdjW < 0)
			tw <<= -ScaleAdjW;
		if (ScaleAdjH < 0)
			th <<= -ScaleAdjH;

		if (MaxW > 512)
			tw = (tw * 512) / MaxW;
		if (MaxH > 512)
			th = (th * 512) / MaxH;

//		fastScaleImage(t1_p->Width, t1_p->Height,(unsigned char*)tsrc,
//							tw, th, (unsigned char*)tdst);

//		gluScaleImage_asm386optimized( GL_RGBA, GL_FALSE, 1,t0_p->Width, t0_p->Height,(unsigned char*)tsrc,
//							tw, th,(unsigned char*)tdst);

		gluScaleImage(GL_RGBA,t1_p->Width, t1_p->Height,GL_UNSIGNED_BYTE,(unsigned char*)tsrc,
							tw, th,GL_UNSIGNED_BYTE, (unsigned char*)tdst);

		if (t1_p->mirrors)
		{
			TexMirrorS(tw, th,t1bfr);
			tw *= 2;
		}

		if (t1_p->mirrort)
		{
			TexMirrorT(tw, th,t1bfr);
			th *= 2;
		}

		if (mt1w > 1)
		{
			if(t1_p->clamps)
				TexClampRepeatS(mt1w,tw, th,t1bfr);
			else
				TexRepeatS(mt1w,tw, th,t1bfr);
			tw *= mt1w;
			aTilew = 0;
		}

		if (mt1h > 1)
		{
			if(t1_p->clampt)
				TexClampRepeatT(mt1h,tw, th,t1bfr);
			else
				TexRepeatT(mt1h,tw, th,t1bfr);
			th *= mt1h;
			aTileh = 0;
		}

		t1width = tw;
		t1height = th;
	}

	realwidth = (t0width >= t1width) ? t0width : t1width;
	realheight = (t0height >= t1height) ? t0height : t1height;

	if (TestPower2(realwidth, realheight, &new_width, &new_height))
	{
		new_width = realwidth;
		new_height = realheight;
	}

	ModTexture(TexMode, realwidth * realheight);

//	fastScaleImage(realwidth, realheight,(unsigned char*)tbuffer1,
//					new_width, new_height, (unsigned char*)tbuffer2);

	realwidth = new_width;
	realheight = new_height;
#define GL_CLAMP_TO_EDGE                  0x812F
	
//	if (t_p->clamps)
	if (tC_p->clamps)
//		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	else
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);

//	if (t_p->clampt)
	if (tC_p->clampt)
//		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	else
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	
//** Build the texture...
	{
		glTexImage2D(GL_TEXTURE_2D, 0, 4, realwidth, realheight, 0, GL_RGBA,
//				 GL_UNSIGNED_BYTE, (unsigned char*)tbuffer2);
				 GL_UNSIGNED_BYTE, (unsigned char*)tbuffer1);
	}

//** is needed for a lot of demos (FiRES) - if the combines mode work right
//** we dont need this anymore

	num_cached_texture++;

	p->info.aTilew = aTilew;
	p->info.aTileh = aTileh;
	p->info.Multw = Multw;
	p->info.Multh = Multh;

	TexMode = 0;
	TexColor = 0;
	return 1;
}

/*****************************************************************************/


static void NCopyTextureRGBA32(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
	_s32 x, y;
	_u8 *dest = i_dst;
    _u8 *src = i_src;

    int xc, yc;
    int xl, yl;
    int dx, dy;

	int flip;

	_s32 twidth = txwidth / 4;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
		x = sx;
   		src = i_src + tstart + ((y * twidth + x) << 2);

	    if ((y & 1) && (swapped))
			flip = 0x04; 
		else 
			flip = 0x00;

		for(xc = 0; xc <= xl; xc++)
		{
			(_u32)src ^= flip; // wtf am I doing here?
			src += 3;
            *dest = *(_u8*)((_u32)src ^ flip); src--; dest++;
            *dest = *(_u8*)((_u32)src ^ flip); src--; dest++;
            *dest = *(_u8*)((_u32)src ^ flip); src--; dest++;
            *dest = *(_u8*)((_u32)src ^ flip); dest++;
            src += 4;
			//x += dx;
        }
		y += dy;
	}


}

// correct !!

static void NCopyTextureRGBA16(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
	_s32 x, y;
	_u8 *dest = i_dst;
    _u16 *src = (_u16*)i_src;

    int xc, yc;
    int xl, yl;
    int dx, dy;
	int flip;

	_s32 twidth = txwidth / 2;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
		_u32  src_l = tstart + ((y * twidth + sx) << 1);

	    if ((y & 1) && (swapped))
			flip = 0x06;
		else
			flip = 0x02;

		x = sx + (tstart >> 1);
		for(xc = 0; xc <= xl; xc++)
           {
			//_u32  src_l = (_u32)i_src + ((y * twidth + x) << 1);
			_u16 src_h = *(_u16 *)((_u32)i_src + (src_l ^flip));

			*(dest) = (((src_h >> 11 ) & 0x001f)<<3); dest++;
			*(dest) = (((src_h >>  6 ) & 0x001f)<<3); dest++;
			*(dest) = (((src_h >>  1 ) & 0x001f)<<3); dest++;
			*(dest) =   (src_h & 0x01) ? 0xff : 0x00; dest++;
//			src++;
			src_l += 0x02;
			//x += dx;
         }
		y += dy;
	}
}

static void NCopyTextureCI8(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
   // tested //
	_u32 *d = (_u32*)i_dst;
	_u8  *s = i_src;
	unsigned int     x, y;
	_u16  *pal = (_u16 *)((_u8 *)(pRDRAM + rdp_reg.tlut_8_addr));   /* palette */
	_u8 index;
	_u32 tsrc;
	_s32 twidth = txwidth;

	int flip;

    int xc, yc;
    int xl, yl;
    int dx, dy;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	y = sy;

	if (pltmode != 0x0000c000)
		Palette8=&Palette8RGBA[0];
	else
		Palette8=&Palette8IA[0];

	for(yc = 0; yc <= yl; yc++)
	{
	    if ((y & 1) && (swapped))
			flip = 0x07;
		else
			flip = 0x03;

		x = sx + tstart;
		tsrc = ((y * twidth) + x);
		for(xc = 0; xc <= xl; xc++)
		{
			s = (_u8 *)(((_u32)i_src) + (tsrc ^ flip));
			index = *s ^0x01;
			*d++ = Palette8[index];
			tsrc++;
		}
		y += dy;
	}
}


static void NCopyTextureCI4(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
	// TT
    _u32 x, y, tsrc;
    int xc, yc;
    int xl, yl;
    int dx, dy;
    _u32 *d = (_u32*)i_dst;
    _u8 *s = i_src;
    _u32 color;

	_u32 *palette;// = (_u32 *)(&Palette8)+(t_p->palette<<4);   /* palette */
	_s32 twidth = txwidth * 2;

    _u8 index;
	int tw;
	int flip;

	if (pltmode != 0x0000c000)
		palette=&Palette8RGBA[(tp_t->palette<<4)&0xf0];
	else
		palette=&Palette8IA[(tp_t->palette<<4)&0xf0];

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
	    if ((y & 1) && (swapped))
			flip = 0x07;
		else
			flip = 0x03;

		x = sx + (tstart<<1);
		tw = (y * twidth + x);
		tsrc = ((tw >> 1));
		for(xc = 0; xc <= xl; xc += 2)
		{
			s = (_u8 *)((_u32)i_src + (tsrc ^ flip));

            index = *s ^0x11;

			color = palette[(index >> 4) & 0x0f];
			*d++ = color;

			color = palette[index & 0x0f];
			*d++ = color;

			tsrc++;
			x += dx;
        }
		if (!(xl & 1)) d--;
		y += dy;
	}
}

static void NCopyTextureIA16(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
	// TT
	_u32 x, y, tsrc;
	_u8 *dest = i_dst;
    _u8 *src = i_src;
	_u8 intensity;

	_s32 twidth = txwidth / 2;

    int xc, yc;
    int xl, yl;
    int dx, dy;
	int flip;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
	    if ((y & 1) && (swapped)) flip = 0x06; else flip = 0x02;
		x = (y * twidth) + sx + (tstart >> 1);
		for(xc = 0; xc <= xl; xc++)
                {
			tsrc = (_u32)i_src + ((x << 1) ^ flip);
			src = (_u8 *)tsrc;
			intensity = *(src + 1);
			*dest =  intensity; dest++;
			*dest =  intensity; dest++;
			*dest =  intensity; dest++;
 	  	    *dest = *(src + 0); dest++;
			x += dx;
        }
		y += dy;
	}
}

static void NCopyTextureIA8(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
	// TT
	_u32 x,y,i,tsrc;
	_u8 *dest = i_dst;
    _u8 *src = i_src;

    int xc, yc;
    int xl, yl;
    int dx, dy;

	_s32 twidth = txwidth;
	int flip;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	tstart = tp_t->offset;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
	    if ((y & 1) && (swapped))
			flip = 0x07;
		else
			flip = 0x03;

		x = (y * twidth) + sx + tstart;

		for(xc = 0; xc <= xl; xc++)
		{
			tsrc = (_u32)i_src + (x ^ flip);
			src = (_u8 *)tsrc;
			i = (*src &0xf0);
			i |= i >> 4;
			*dest = (_u8)i; dest++;
			*dest = (_u8)i; dest++;
			*dest = (_u8)i; dest++;
 	  	    *dest = (*src & 0x0f) | ((*src & 0x0f) << 4); dest++; // aplha
			x += dx;
        }
		y += dy;
	}
}

static void NCopyTextureIA4(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
	// TT
    _u32 x, y, tsrc;
    _u8 *d = i_dst;
    _u8 *s = i_src;
    _u8 i;
	int tw;


    int xc, yc;
    int xl, yl;
    int dx, dy;

	_s32 twidth = txwidth * 2;
	int flip;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
	    if ((y & 1) && (swapped))
			flip = 0x07;
		else
			flip = 0x03;

		x = (y * twidth) + sx + (tstart << 1);
		tw = x;

		for(xc = 0; xc <= xl; xc += 2)
		{
			/* 0x800 might be not correct! */
			tsrc = (_u32)i_src + ((tw >> 1)^ flip);
			s = (_u8 *)tsrc;

			i = (*s & 0xf0) | ((*s & 0xf0) >> 4);
			*d++ = i;
			*d++ = i;
			*d++ = i;
			*d++ = (i & 1) ? 0xff : 0x00;

			i = (*s & 0x0f) | ((*s & 0x0f) << 4);
			*d++ = i;
			*d++ = i;
			*d++ = i;
			*d++ = (i & 1) ? 0xff : 0x00;

			tw += 2;
//			x += dx;
        }
		if (!(xl & 1)) d-=4;
		y += dy;
	}
}

static void NCopyTextureI16(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
	_u32 x, y, tsrc;
	_u8 *dest = i_dst;
    _u8 *src = i_src;
	_u8 intensity;

	_s32 twidth = txwidth / 2;

    int xc, yc;
    int xl, yl;
    int dx, dy;
	int flip;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
	    if ((y & 1) && (swapped))
			flip = 0x06;
		else
			flip = 0x02;

		x = (y * twidth) + sx + (tstart >> 1);
		for(xc = 0; xc <= xl; xc++)
        {
			tsrc = (_u32)i_src + ((x << 1) ^ flip);
			src = (_u8 *)tsrc;
			intensity = *(src + 1);
			*dest =  intensity; dest++;
			*dest =  intensity; dest++;
			*dest =  intensity; dest++;
 	  	    *dest = *(src + 1); dest++;
			x += dx;
        }
		y += dy;
	}
}

static void NCopyTextureI8(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
        _u32  x, y;
        _u8  *d = i_dst;
        _u8  *s = i_src;
        _u8  i;


    int xc, yc;
    int xl, yl;
    int dx, dy;

	_s32 twidth = txwidth;
	int flip;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	tstart = tp_t->offset;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
	    if ((y & 1) && (swapped))
			flip = 0x07;
		else
			flip = 0x03;

		x = (y * twidth) + sx + tstart;

		for(xc = 0; xc <= xl; xc++)
        {
                    s = i_src + (x ^ flip);
					i = *s;
					*d++ = i;
                    *d++ = i;
                    *d++ = i;
                    *d++ = i; 
                    //s++;
					x += dx;
        }
		y += dy;
	}
}

static void NCopyTextureI4(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _s32 sx, _s32 ex, _s32 sy, _s32 ey)
{
	// TT
    _u32 x, y, tsrc;
    _u8 *d = i_dst;
    _u8 *s = i_src;
    _u8 i;
	int tw;
	int flip;

	_s32 twidth = txwidth * 2;

    int xc, yc;
    int xl, yl;
    int dx, dy;

	yl = ey - sy;
	xl = ex - sx;

	dy = 1;
	dx = 1;

	y = sy;

	for(yc = 0; yc <= yl; yc++)
	{
	    if ((y & 1) && (swapped))
			flip = 0x07;
		else
			flip = 0x03;

		x = (y * twidth) + sx + (tstart << 1);
		tw = x;

		for(xc = 0; xc <= xl; xc += 2)
        {
			tsrc = (_u32)i_src + ((tw >> 1)^ flip);
			s = (_u8 *)tsrc;

			i = (*s & 0xf0) | ((*s & 0xf0) >> 4);
			*d++ = i;
			*d++ = i;
			*d++ = i;
			*d++ = i;

			i = (*s & 0x0f) | ((*s & 0x0f) << 4);
			*d++ = i;
			*d++ = i;
			*d++ = i;
			*d++ = i;

			tw += 2;
//			x += dx;
        }
		if (!(xl & 1)) d-=4;
		y += dy;
		//y++;
	}
}


/*****************************************************************************/
/*** the follow function is  not implemented yet *****************************/
/*****************************************************************************/

static void CopyTextureYUV(t_tile *tp_t, _u8 *i_src, _u8 *i_dst, _u32 width, _u32 height)
{
    _u32 i, j;
    _u8 *d = i_dst;
    _u16 *s = (_u16*)i_src;

    for(i=0; i < height; i++)
        for(j=0; j < width; j++)
        {
            *d++ = 100;
            *d++ = 150;
            *d++ = 200;
            *d++ = 255;
            s++;
        }

}
